"""User module."""

from typing import Optional, cast

from wriftai._resource import Resource
from wriftai.common_types import NotRequired, StrEnum, User
from wriftai.pagination import PaginatedResponse, PaginationOptions


class UsersSortBy(StrEnum):
    """Enumeration of possible sorting options for querying users."""

    CREATED_AT_ASC = "created_at_asc"
    CREATED_AT_DESC = "created_at_desc"


class UserPaginationOptions(PaginationOptions):
    """Pagination options for querying users."""

    sort_by: NotRequired[UsersSortBy]
    """The sorting criteria."""


class UsersResource(Resource):
    """Resource for operations related to users."""

    _API_PREFIX = "/users"
    _SEARCH_USERS_SUFFIX = "/users"

    def get(self, username: str) -> User:
        """Fetch a user by their username.

        Args:
            username(str): The username of the user.

        Returns:
            User: The user object.
        """
        response = self._api.request("GET", f"{self._API_PREFIX}/{username}")
        return cast(User, response)

    async def async_get(self, username: str) -> User:
        """Fetch a user by their username.

        Args:
            username(str): The username of the user.

        Returns:
            User: The user object.
        """
        response = await self._api.async_request(
            "GET", f"{self._API_PREFIX}/{username}"
        )
        return cast(User, response)

    def list(
        self, pagination_options: Optional[UserPaginationOptions] = None
    ) -> PaginatedResponse[User]:
        """List users.

        Args:
            pagination_options (Optional[PaginationOptions]): Optional settings
                to control pagination behavior.

        Returns:
            PaginatedResponse[User]: Paginated response containing users
                and navigation metadata.
        """
        response = self._api.request(
            method="GET", params=pagination_options, path=self._API_PREFIX
        )

        # The response will always match the PaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]

    async def async_list(
        self, pagination_options: Optional[UserPaginationOptions] = None
    ) -> PaginatedResponse[User]:
        """List users.

        Args:
            pagination_options (Optional[PaginationOptions]): Optional settings
                to control pagination behavior.

        Returns:
            PaginatedResponse[User]: Paginated response containing users
                and navigation metadata.
        """
        response = await self._api.async_request(
            method="GET", params=pagination_options, path=self._API_PREFIX
        )
        # The response will always match the PaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]

    def search(
        self, q: str, pagination_options: Optional[PaginationOptions] = None
    ) -> PaginatedResponse[User]:
        """Search users.

        Args:
            q (str): The search query.
            pagination_options (Optional[PaginationOptions]): Optional settings to
                control pagination behavior.

        Returns:
            PaginatedResponse[User]: Paginated response containing users
                and navigation metadata.
        """
        params = self._build_search_params(q, pagination_options)

        response = self._api.request(
            method="GET",
            params=params,
            path=f"{self._SEARCH_API_PREFIX}{self._SEARCH_USERS_SUFFIX}",
        )
        # The response will always match the PaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]

    async def async_search(
        self, q: str, pagination_options: Optional[PaginationOptions] = None
    ) -> PaginatedResponse[User]:
        """Search Users.

        Args:
            q (str): The search query.
            pagination_options (Optional[PaginationOptions]): Optional settings to
                control pagintation behavior.

        Returns:
            PaginatedResponse[User]: Paginated response containing users
                and navigation metadata.
        """
        params = self._build_search_params(q, pagination_options)

        response = await self._api.async_request(
            method="GET",
            params=params,
            path=f"{self._SEARCH_API_PREFIX}{self._SEARCH_USERS_SUFFIX}",
        )
        # The response will always match the PaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]



from typing import Optional
from unittest.mock import AsyncMock, Mock, patch

import pytest

from wriftai.models import (
    CreateModelParams,
    ModelPaginationOptions,
    ModelsResource,
    ModelsSortBy,
    ModelVisibility,
    UpdateModelParams,
)
from wriftai.pagination import PaginationOptions


def test_delete() -> None:
    mock_api = Mock()
    mock_api.request = Mock()

    model = ModelsResource(api=mock_api)
    test_owner = "test_user"
    test_model_name = "dummy_model"

    model.delete(owner=test_owner, name=test_model_name)

    mock_api.request.assert_called_once_with(
        "DELETE", f"{model._MODELS_API_PREFIX}/{test_owner}/{test_model_name}"
    )


@pytest.mark.asyncio
async def test_async_delete() -> None:
    mock_api = Mock()
    mock_api.async_request = AsyncMock()

    model = ModelsResource(api=mock_api)
    test_owner = "test_user"
    test_model_name = "dummy_model"

    await model.async_delete(owner=test_owner, name=test_model_name)

    mock_api.async_request.assert_called_once_with(
        "DELETE", f"{model._MODELS_API_PREFIX}/{test_owner}/{test_model_name}"
    )


@patch("wriftai.models.PaginatedResponse")
@pytest.mark.parametrize("owner", [None, "test_user"])
def test_list(mock_paginated_response: Mock, owner: Optional[str]) -> None:
    mock_api = Mock()
    test_response = {"key": "value"}
    mock_api.request.return_value = test_response

    models = ModelsResource(api=mock_api)
    pagination_options = ModelPaginationOptions({
        "cursor": "abc123",
        "page_size": 50,
        "sort_by": ModelsSortBy.CREATED_AT_ASC,
    })

    path = (
        models._MODELS_API_PREFIX
        if owner is None
        else f"{models._MODELS_API_PREFIX}/{owner}"
    )

    result = models.list(pagination_options=pagination_options, owner=owner)

    mock_api.request.assert_called_once_with(
        method="GET", path=path, params=pagination_options
    )
    mock_paginated_response.assert_called_once_with(**test_response)
    assert result == mock_paginated_response.return_value


@patch("wriftai.models.PaginatedResponse")
@pytest.mark.parametrize("owner", [None, "test_user"])
@pytest.mark.asyncio
async def test_async_list(mock_paginated_response: Mock, owner: Optional[str]) -> None:
    mock_api = AsyncMock()
    test_response = {"key": "value"}
    mock_api.async_request.return_value = test_response

    models = ModelsResource(api=mock_api)
    pagination_options = ModelPaginationOptions({
        "cursor": "abc123",
        "page_size": 50,
        "sort_by": ModelsSortBy.CREATED_AT_ASC,
    })
    path = (
        models._MODELS_API_PREFIX
        if owner is None
        else f"{models._MODELS_API_PREFIX}/{owner}"
    )

    result = await models.async_list(pagination_options=pagination_options, owner=owner)

    mock_api.async_request.assert_called_once_with(
        method="GET", path=path, params=pagination_options
    )
    mock_paginated_response.assert_called_once_with(**test_response)
    assert result == mock_paginated_response.return_value


def test_get() -> None:
    mock_api = Mock()
    mock_api.request = Mock()

    models = ModelsResource(api=mock_api)
    test_owner = "test_user"
    test_model_name = "dummy_model"

    result = models.get(owner=test_owner, name=test_model_name)

    mock_api.request.assert_called_once_with(
        "GET", f"{models._MODELS_API_PREFIX}/{test_owner}/{test_model_name}"
    )

    assert result == mock_api.request.return_value


@pytest.mark.asyncio
async def test_async_get() -> None:
    mock_api = Mock()
    mock_api.async_request = AsyncMock()

    models = ModelsResource(api=mock_api)
    test_owner = "test_user"
    test_model_name = "dummy_model"

    result = await models.async_get(owner=test_owner, name=test_model_name)

    mock_api.async_request.assert_called_once_with(
        "GET", f"{models._MODELS_API_PREFIX}/{test_owner}/{test_model_name}"
    )

    assert result == mock_api.async_request.return_value


def test_create() -> None:
    mock_api = Mock()
    mock_api.request = Mock()

    models = ModelsResource(api=mock_api)

    model_data: CreateModelParams = {
        "name": "test_model",
        "visibility": ModelVisibility.public,
        "hardware_name": "NVIDIA A100",
        "description": "Test model description",
        "source_url": "https://example.com/source",
        "license_url": "https://example.com/license",
        "paper_url": "https://example.com/paper",
    }

    result = models.create(model_data)

    mock_api.request.assert_called_once_with(
        method="POST",
        path=models._MODELS_API_PREFIX,
        body=model_data,
    )

    assert result == mock_api.request.return_value


@pytest.mark.asyncio
async def test_async_create() -> None:
    mock_api = Mock()
    mock_api.async_request = AsyncMock()

    models = ModelsResource(api=mock_api)

    model_data: CreateModelParams = {
        "name": "test_model",
        "visibility": ModelVisibility.public,
        "hardware_name": "NVIDIA A100",
        "description": "Test model description",
        "source_url": "https://example.com/source",
        "license_url": "https://example.com/license",
        "paper_url": "https://example.com/paper",
    }

    result = await models.async_create(model_data)

    mock_api.async_request.assert_called_once_with(
        method="POST",
        path=models._MODELS_API_PREFIX,
        body=model_data,
    )

    assert result == mock_api.async_request.return_value


def test_update() -> None:
    mock_api = Mock()
    models = ModelsResource(api=mock_api)
    test_owner = "test_user"
    test_model_name = "dummy_model"
    payload: UpdateModelParams = {
        "name": "updated_model_name",
        "description": "Updated description",
        "source_url": "https://example.com/updated_source",
        "license_url": "https://example.com/updated_license",
        "paper_url": "https://example.com/updated_paper",
        "hardware_name": "Updated Hardware",
        "visibility": ModelVisibility.public,
    }

    result = models.update(
        owner=test_owner,
        name=test_model_name,
        params=payload,
    )

    mock_api.request.assert_called_once_with(
        method="PATCH",
        path=f"{models._MODELS_API_PREFIX}/{test_owner}/{test_model_name}",
        body=payload,
    )

    assert result == mock_api.request.return_value


@pytest.mark.asyncio
async def test_async_update() -> None:
    mock_api = AsyncMock()
    models = ModelsResource(api=mock_api)
    test_owner = "test_user"
    test_model_name = "dummy_model"
    payload: UpdateModelParams = {
        "name": "updated_model_name",
        "description": "Updated description",
        "visibility": ModelVisibility.public,
    }

    result = await models.async_update(
        owner=test_owner,
        name=test_model_name,
        params=payload,
    )

    mock_api.async_request.assert_called_once_with(
        method="PATCH",
        path=f"{models._MODELS_API_PREFIX}/{test_owner}/{test_model_name}",
        body=payload,
    )

    assert result == mock_api.async_request.return_value


@patch("wriftai.models.PaginatedResponse")
def test_search(mock_paginated_response: Mock) -> None:
    mock_api = Mock()
    test_response = {"key": "value"}
    mock_api.request.return_value = test_response

    models = ModelsResource(api=mock_api)
    pagination_options = PaginationOptions({
        "cursor": "abc123",
        "page_size": 50,
    })
    result = models.search(q="dummy_query", pagination_options=pagination_options)

    expected_params = {
        "cursor": pagination_options["cursor"],
        "page_size": pagination_options["page_size"],
        "q": "dummy_query",
    }

    mock_api.request.assert_called_once_with(
        method="GET",
        path=f"{models._SEARCH_API_PREFIX}{models._SEARCH_MODELS_SUFFIX}",
        params=expected_params,
    )
    mock_paginated_response.assert_called_once_with(**test_response)
    assert result == mock_paginated_response.return_value


@patch("wriftai.models.PaginatedResponse")
@pytest.mark.asyncio
async def test_async_search(mock_paginated_response: Mock) -> None:
    mock_api = AsyncMock()
    test_response = {"key": "value"}
    mock_api.async_request.return_value = test_response

    models = ModelsResource(api=mock_api)
    pagination_options = PaginationOptions({
        "cursor": "abc123",
        "page_size": 50,
    })
    result = await models.async_search(
        q="dummy_query", pagination_options=pagination_options
    )

    expected_params = {
        "cursor": pagination_options["cursor"],
        "page_size": pagination_options["page_size"],
        "q": "dummy_query",
    }

    mock_api.async_request.assert_called_once_with(
        method="GET",
        path=f"{models._SEARCH_API_PREFIX}{models._SEARCH_MODELS_SUFFIX}",
        params=expected_params,
    )
    mock_paginated_response.assert_called_once_with(**test_response)
    assert result == mock_paginated_response.return_value



"""Package initializer for WriftAI Python Client."""

from wriftai._client import Client, ClientOptions
from wriftai.pagination import PaginationOptions

__all__ = ["Client", "ClientOptions", "PaginationOptions"]



"""Authenticated user module."""

from typing import Optional, TypedDict, cast

from wriftai._resource import Resource
from wriftai.common_types import NotRequired, User
from wriftai.models import Model
from wriftai.pagination import PaginatedResponse, PaginationOptions


class UpdateUserParams(TypedDict):
    """Parameters for updating a user."""

    username: NotRequired[str]
    """The new username of the user."""
    name: NotRequired[str | None]
    """The new name of the user."""
    bio: NotRequired[str | None]
    """The new biography of the user."""
    urls: NotRequired[list[str] | None]
    """The new URLs associated with the user."""
    company: NotRequired[str | None]
    """The new company of the user."""
    location: NotRequired[str | None]
    """The new location of the user."""


class AuthenticatedUser(Resource):
    """Resource for operations related to the authenticated user."""

    _API_PREFIX = "/user"

    def get(self) -> User:
        """Get the authenticated user.

        Returns:
            User: The user object.
        """
        response = self._api.request("GET", f"{self._API_PREFIX}")
        return cast(User, response)

    async def async_get(self) -> User:
        """Get authenticated user.

        Returns:
            User: The user object.
        """
        response = await self._api.async_request("GET", f"{self._API_PREFIX}")
        return cast(User, response)

    def models(
        self, pagination_options: Optional[PaginationOptions] = None
    ) -> PaginatedResponse[Model]:
        """List models of the authenticated user.

        Args:
            pagination_options (Optional[PaginationOptions]): Optional settings
                to control pagination behavior.

        Returns:
            PaginatedResponse[Model]: Paginated response containing models
                and navigation metadata.
        """
        response = self._api.request(
            method="GET",
            params=pagination_options,
            path=f"{self._API_PREFIX}{self._MODELS_API_PREFIX}",
        )
        # The response will always match the PaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]

    async def async_models(
        self, pagination_options: Optional[PaginationOptions] = None
    ) -> PaginatedResponse[Model]:
        """List models of the authenticated user.

        Args:
            pagination_options (Optional[PaginationOptions]): Optional settings
                to control pagination behavior.

        Returns:
            PaginatedResponse[Model]: Paginated response containing models
                and navigation metadata.
        """
        response = await self._api.async_request(
            method="GET",
            params=pagination_options,
            path=f"{self._API_PREFIX}{self._MODELS_API_PREFIX}",
        )
        # The response will always match the PaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]

    def update(
        self,
        params: UpdateUserParams,
    ) -> User:
        """Update the authenticated user.

        Args:
            params (UpdateUserParams): The fields to update.

        Returns:
            User: The updated user object.
        """
        response = self._api.request(
            method="PATCH",
            path=f"{self._API_PREFIX}",
            # The params matches JsonValue at runtime,
            # but static type checkers may not infer this correctly.
            # Hence, we ignore the argument type warning.
            body=params,  # type:ignore[arg-type]
        )
        return cast(User, response)

    async def async_update(
        self,
        params: UpdateUserParams,
    ) -> User:
        """Update the authenticated user.

        Args:
            params (UpdateUserParams): The fields to update.

        Returns:
            User: The updated user object.
        """
        response = await self._api.async_request(
            method="PATCH",
            path=f"{self._API_PREFIX}",
            # The params matches JsonValue at runtime,
            # but static type checkers may not infer this correctly.
            # Hence, we ignore the argument type warning.
            body=params,  # type:ignore[arg-type]
        )
        return cast(User, response)



from unittest.mock import AsyncMock, Mock, patch

import pytest

from wriftai.pagination import PaginationOptions
from wriftai.users import UserPaginationOptions, UsersResource, UsersSortBy


def test_get() -> None:
    mock_api = Mock()
    mock_api.request = Mock()

    user = UsersResource(api=mock_api)
    username = "test_username"

    result = user.get(username=username)

    mock_api.request.assert_called_once_with("GET", f"{user._API_PREFIX}/{username}")
    assert result == mock_api.request.return_value


@pytest.mark.asyncio
async def test_async_get() -> None:
    mock_api = Mock()
    mock_api.async_request = AsyncMock()

    user = UsersResource(api=mock_api)
    username = "test_username"

    result = await user.async_get(username=username)

    mock_api.async_request.assert_called_once_with(
        "GET", f"{user._API_PREFIX}/{username}"
    )
    assert result == mock_api.async_request.return_value


@patch("wriftai.users.PaginatedResponse")
def test_list(mock_paginated_response: Mock) -> None:
    mock_api = Mock()
    test_response = {"key": "value"}
    mock_api.request.return_value = test_response

    users = UsersResource(api=mock_api)
    pagination_options = UserPaginationOptions({
        "cursor": "abc123",
        "page_size": 50,
        "sort_by": UsersSortBy.CREATED_AT_DESC,
    })
    result = users.list(pagination_options=pagination_options)

    mock_api.request.assert_called_once_with(
        method="GET", path=users._API_PREFIX, params=pagination_options
    )
    mock_paginated_response.assert_called_once_with(**test_response)
    assert result == mock_paginated_response.return_value


@patch("wriftai.users.PaginatedResponse")
@pytest.mark.asyncio
async def test_async_list(mock_paginated_response: Mock) -> None:
    mock_api = AsyncMock()
    test_response = {"key": "value"}
    mock_api.async_request.return_value = test_response

    users = UsersResource(api=mock_api)
    pagination_options = UserPaginationOptions({
        "cursor": "abc123",
        "page_size": 50,
        "sort_by": UsersSortBy.CREATED_AT_DESC,
    })
    result = await users.async_list(pagination_options=pagination_options)

    mock_api.async_request.assert_called_once_with(
        method="GET", path=users._API_PREFIX, params=pagination_options
    )
    mock_paginated_response.assert_called_once_with(**test_response)
    assert result == mock_paginated_response.return_value


@patch("wriftai.users.PaginatedResponse")
def test_search(mock_paginated_response: Mock) -> None:
    mock_api = Mock()
    test_response = {"key": "value"}
    mock_api.request.return_value = test_response

    users = UsersResource(api=mock_api)
    pagination_options = PaginationOptions({
        "cursor": "abc123",
        "page_size": 50,
    })
    result = users.search(q="dummy_query", pagination_options=pagination_options)

    expected_params = {
        "cursor": pagination_options["cursor"],
        "page_size": pagination_options["page_size"],
        "q": "dummy_query",
    }

    mock_api.request.assert_called_once_with(
        method="GET",
        path=f"{users._SEARCH_API_PREFIX}{users._SEARCH_USERS_SUFFIX}",
        params=expected_params,
    )
    mock_paginated_response.assert_called_once_with(**test_response)
    assert result == mock_paginated_response.return_value


@patch("wriftai.users.PaginatedResponse")
@pytest.mark.asyncio
async def test_async_search(mock_paginated_response: Mock) -> None:
    mock_api = AsyncMock()
    test_response = {"key": "value"}
    mock_api.async_request.return_value = test_response

    users = UsersResource(api=mock_api)
    pagination_options = PaginationOptions({
        "cursor": "abc123",
        "page_size": 50,
    })
    result = await users.async_search(
        q="dummy_query", pagination_options=pagination_options
    )

    expected_params = {
        "cursor": pagination_options["cursor"],
        "page_size": pagination_options["page_size"],
        "q": "dummy_query",
    }

    mock_api.async_request.assert_called_once_with(
        method="GET",
        path=f"{users._SEARCH_API_PREFIX}{users._SEARCH_USERS_SUFFIX}",
        params=expected_params,
    )
    mock_paginated_response.assert_called_once_with(**test_response)
    assert result == mock_paginated_response.return_value



"""Pagination module."""

from dataclasses import dataclass
from typing import Generic, TypedDict, TypeVar

from wriftai.common_types import NotRequired

T = TypeVar("T")


# Only expose PaginatedResponse to avoid duplicate references in Sphinx docs.
# Including PaginationOptions here will cause it to appear both as
# wriftai.PaginationOptions and wriftai.pagination.PaginationOptions,
# leading to cross-reference warning
__all__ = ["PaginatedResponse"]


class PaginationOptions(TypedDict):
    """Options for pagination."""

    cursor: NotRequired[str]
    """Cursor for pagination."""
    page_size: NotRequired[int]
    """Number of items per page."""


@dataclass
class PaginatedResponse(Generic[T]):
    """Represents a paginated response."""

    items: list[T]
    """List of items returned in the current page."""
    next_cursor: str | None
    """Cursor pointing to the next page."""
    previous_cursor: str | None
    """Cursor pointing to the previous page."""
    next_url: str | None
    """URL to fetch the next page."""
    previous_url: str | None
    """URL to fetch the previous page."""



from unittest.mock import AsyncMock, Mock, patch

import pytest

from wriftai.pagination import PaginationOptions
from wriftai.versions import CreateVersionParams, Versions


def test_get() -> None:
    mock_api = Mock()

    model = Versions(api=mock_api)
    version_id = "c12258c4-ed83-4d7b-a784-8ed55412325a"

    result = model.get(version_id=version_id)

    mock_api.request.assert_called_once_with(
        "GET", f"{model._VERSIONS_API_PREFIX}/{version_id}"
    )
    assert result == mock_api.request.return_value


@pytest.mark.asyncio
async def test_async_get() -> None:
    mock_api = Mock()
    mock_api.async_request = AsyncMock()

    model = Versions(api=mock_api)
    version_id = "c12258c4-ed83-4d7b-a784-8ed55412325a"

    result = await model.async_get(version_id=version_id)

    mock_api.async_request.assert_called_once_with(
        "GET", f"{model._VERSIONS_API_PREFIX}/{version_id}"
    )
    assert result == mock_api.async_request.return_value


@patch("wriftai.versions.PaginatedResponse")
def test_list(mock_paginated_response: Mock) -> None:
    mock_api = Mock()
    test_response = {"key": "value"}
    mock_api.request.return_value = test_response

    versions = Versions(api=mock_api)
    pagination_options = PaginationOptions({"cursor": "abc123", "page_size": 50})
    model_owner = "abc"
    model_name = "textgenerator"

    result = versions.list(
        model_owner=model_owner,
        model_name=model_name,
        pagination_options=pagination_options,
    )
    path = (
        f"{versions._MODELS_API_PREFIX}/{model_owner}"
        f"/{model_name}{versions._VERSIONS_API_SUFFIX}"
    )
    mock_api.request.assert_called_once_with(
        method="GET", path=path, params=pagination_options
    )
    mock_paginated_response.assert_called_once_with(**test_response)
    assert result == mock_paginated_response.return_value


@patch("wriftai.versions.PaginatedResponse")
@pytest.mark.asyncio
async def test_async_list(mock_paginated_response: Mock) -> None:
    mock_api = AsyncMock()
    test_response = {"key": "value"}
    mock_api.async_request.return_value = test_response

    versions = Versions(api=mock_api)
    pagination_options = PaginationOptions({"cursor": "abc123", "page_size": 50})
    model_owner = "abc"
    model_name = "textgenerator"

    result = await versions.async_list(
        model_owner=model_owner,
        model_name=model_name,
        pagination_options=pagination_options,
    )

    path = (
        f"{versions._MODELS_API_PREFIX}/{model_owner}"
        f"/{model_name}{versions._VERSIONS_API_SUFFIX}"
    )
    mock_api.async_request.assert_called_once_with(
        method="GET", path=path, params=pagination_options
    )
    mock_paginated_response.assert_called_once_with(**test_response)
    assert result == mock_paginated_response.return_value


def test_delete() -> None:
    mock_api = Mock()

    model = Versions(api=mock_api)
    version_id = "c12258c4-ed83-4d7b-a784-8ed55412325a"

    model.delete(version_id=version_id)

    mock_api.request.assert_called_once_with(
        "DELETE", f"{model._VERSIONS_API_PREFIX}/{version_id}"
    )


@pytest.mark.asyncio
async def test_async_delete() -> None:
    mock_api = Mock()
    mock_api.async_request = AsyncMock()

    model = Versions(api=mock_api)
    version_id = "c12258c4-ed83-4d7b-a784-8ed55412325a"

    await model.async_delete(version_id=version_id)

    mock_api.async_request.assert_called_once_with(
        "DELETE", f"{model._VERSIONS_API_PREFIX}/{version_id}"
    )


def test_create() -> None:
    mock_api = Mock()

    model = Versions(api=mock_api)

    model_owner = "abc"
    model_name = "textgenerator"
    options: CreateVersionParams = {
        "release_notes": "Initial release with basic features",
        "container_image_digest": "sha256:" + "a" * 64,
        "schemas": {
            "prediction": {
                "input": {"key1": "value1", "key2": 123},
                "output": {"result": True, "message": "Success"},
            }
        },
    }

    result = model.create(
        model_owner=model_owner,
        model_name=model_name,
        options=options,
    )
    path = (
        f"{model._MODELS_API_PREFIX}/{model_owner}"
        f"/{model_name}{model._VERSIONS_API_SUFFIX}"
    )
    mock_api.request.assert_called_once_with(
        "POST",
        path,
        body=options,
    )

    assert result == mock_api.request.return_value


@pytest.mark.asyncio
async def test_async_create() -> None:
    mock_api = Mock()
    mock_api.async_request = AsyncMock()

    model = Versions(api=mock_api)

    model_owner = "abc"
    model_name = "textgenerator"
    options: CreateVersionParams = {
        "release_notes": "Initial release with basic features",
        "container_image_digest": "sha256:" + "a" * 64,
        "schemas": {
            "prediction": {
                "input": {"key1": "value1", "key2": 123},
                "output": {"result": True, "message": "Success"},
            }
        },
    }

    result = await model.async_create(
        model_owner=model_owner,
        model_name=model_name,
        options=options,
    )

    path = (
        f"{model._MODELS_API_PREFIX}/{model_owner}"
        f"/{model_name}{model._VERSIONS_API_SUFFIX}"
    )

    mock_api.async_request.assert_called_once_with(
        "POST",
        path,
        body=options,
    )

    assert result == mock_api.async_request.return_value



"""API client module."""

from collections.abc import Mapping
from typing import Any, Optional, cast

import httpx

from wriftai.common_types import JsonValue


class API:
    """Client for WriftAI's API."""

    def __init__(
        self, sync_client: httpx.Client, async_client: httpx.AsyncClient
    ) -> None:
        """Initializes the API with synchronous and asynchronous HTTP clients.

        Args:
            sync_client (httpx.Client): An instance of a synchronous HTTP client.
            async_client (httpx.AsyncClient): An instance of an asynchronous HTTP
                client.
        """
        self._sync_client = sync_client
        self._async_client = async_client

    def request(
        self,
        method: str,
        path: str,
        body: Optional[JsonValue] = None,
        headers: Optional[dict[str, Any]] = None,
        params: Optional[Mapping[str, Any]] = None,
    ) -> JsonValue:
        """Sends a synchronous HTTP request using the configured sync client.

        Args:
            method (str): The HTTP method to use (e.g., 'GET', 'POST').
            path (str): The URL path to send the request to.
            body (`Optional[JsonValue]`): The JSON body to include in
                the request.
            headers (`Optional[dict[str, Any]]`): Optional HTTP headers to
                include in the request.
            params (`Optional[Mapping[str, Any]]`): Optional query parameters.


        Returns:
            JsonValue: The json response received from the server.
        """
        response = self._sync_client.request(
            method=method, url=path, json=body, headers=headers, params=params
        )
        response.raise_for_status()
        return cast(JsonValue, response.json())

    async def async_request(
        self,
        method: str,
        path: str,
        body: JsonValue = None,
        headers: Optional[dict[str, Any]] = None,
        params: Optional[Mapping[str, Any]] = None,
    ) -> JsonValue:
        """Sends an asynchronous HTTP request using the configured async client.

        Args:
            method (str): The HTTP method to use (e.g., 'GET', 'POST').
            path (str): The URL path to send the request to.
            body (`Optional[JsonValue]`): The JSON body to include in
                the request.
            headers (`Optional[dict[str, Any]]`): Optional HTTP headers to
                include in the request.
            params (`Optional[Mapping[str, Any]]`): Optional query parameters.


        Returns:
            JsonValue: The json response received from the server.
        """
        response = await self._async_client.request(
            method=method, url=path, json=body, headers=headers, params=params
        )
        response.raise_for_status()
        return cast(JsonValue, response.json())



from copy import copy
from unittest.mock import AsyncMock, Mock, call, patch

import pytest

from wriftai.predictions import (
    CreatePredictionParams,
    Predictions,
    PredictionWithIO,
    Status,
    WaitOptions,
)


def test_get() -> None:
    test_prediction_id = "test_id"
    mock_api = Mock()
    mock_api.request = Mock()

    predictions = Predictions(api=mock_api)

    result = predictions.get(test_prediction_id)

    mock_api.request.assert_called_once_with(
        method="GET", path=f"{predictions._API_PREFIX}/{test_prediction_id}"
    )

    assert result == mock_api.request.return_value


@pytest.mark.asyncio
async def test_async_get() -> None:
    test_prediction_id = "test_id"
    mock_api = Mock()
    mock_api.async_request = AsyncMock()

    predictions = Predictions(api=mock_api)

    result = await predictions.async_get(test_prediction_id)

    mock_api.async_request.assert_awaited_once_with(
        method="GET", path=f"{predictions._API_PREFIX}/{test_prediction_id}"
    )

    assert result == mock_api.async_request.return_value


def test__prediction_path_latest_version() -> None:
    predictions = Predictions(api=Mock())
    owner = "test_owner"
    model = "test_model"
    path = predictions._prediction_path(model_owner=owner, model_name=model)
    expected = (
        f"{predictions._MODELS_API_PREFIX}/{owner}/{model}"
        f"{predictions._PREDICTIONS_API_SUFFIX}"
    )
    assert path == expected


def test__prediction_path_specific_version() -> None:
    predictions = Predictions(api=Mock())
    version_id = "c12258c4-ed83-4d7b-a784-8ed55412325a"
    path = predictions._prediction_path(version_id=version_id)
    expected = (
        f"{predictions._VERSIONS_API_PREFIX}/{version_id}"
        f"{predictions._PREDICTIONS_API_SUFFIX}"
    )
    assert path == expected


def test__prediction_path_invalid_params() -> None:
    predictions = Predictions(api=Mock())
    owner = "test_owner"
    with pytest.raises(TypeError) as exc_info:
        predictions._prediction_path(model_owner=owner)
    assert str(exc_info.value) == predictions._ERROR_MSG_INVALID_PREDICTION_PARAMS


@pytest.mark.parametrize("validate_input", [True, False])
@pytest.mark.parametrize("wait", [True, False])
def test_create(validate_input: bool, wait: bool) -> None:
    mock_api = Mock()
    predictions = Predictions(api=mock_api)

    params: CreatePredictionParams = {
        "input": {"key": "value"},
        "webhook_url": "https://example.com/webhook",
    }

    if validate_input:
        params["validate_input"] = validate_input

    path = "/models/test_owner/test_model/predictions"
    mock_prediction = {"id": "pred-123", "status": "succeeded"}
    wait_options = WaitOptions(poll_interval=2)

    with (
        patch.object(predictions, "_prediction_path", return_value=path) as mock_path,
        patch.object(predictions, "wait", return_value=mock_prediction) as mock_wait,
    ):
        mock_api.request.return_value = mock_prediction

        result = predictions.create(
            model_owner="test_owner",
            model_name="test_model",
            params=params,
            wait=wait,
            wait_options=wait_options,
        )

        headers = {"Validate-Input": "true"} if validate_input else None

        mock_path.assert_called_once_with("test_owner", "test_model", None)

        mock_api.request.assert_called_once_with(
            method="POST",
            path=path,
            body={"input": params["input"], "webhook_url": params["webhook_url"]},
            headers=headers,
        )

        if wait:
            mock_wait.assert_called_once_with(
                mock_prediction["id"], options=wait_options
            )
        else:
            mock_wait.assert_not_called()

        assert result == mock_api.request.return_value


@pytest.mark.asyncio
@pytest.mark.parametrize("validate_input", [True, False])
@pytest.mark.parametrize("wait", [True, False])
async def test_async_create(validate_input: bool, wait: bool) -> None:
    mock_api = Mock()
    mock_api.async_request = AsyncMock()
    predictions = Predictions(api=mock_api)

    params: CreatePredictionParams = {
        "input": {"key": "value"},
        "webhook_url": "https://example.com/webhook",
    }
    if validate_input:
        params["validate_input"] = validate_input

    version_id = "c12258c4-ed83-4d7b-a784-8ed55412325a"
    path = f"versions/{version_id}/predictions"
    mock_prediction = {"id": "pred-123", "status": "succeeded"}
    wait_options = WaitOptions(poll_interval=2)

    with (
        patch.object(predictions, "_prediction_path", return_value=path) as mock_path,
        patch.object(
            predictions, "async_wait", return_value=mock_prediction
        ) as mock_wait,
    ):
        mock_api.async_request.return_value = mock_prediction

        result = await predictions.async_create(
            version_id=version_id, params=params, wait=wait, wait_options=wait_options
        )

        headers = {"Validate-Input": "true"} if validate_input else None

        mock_path.assert_called_once_with(None, None, version_id)

        mock_api.async_request.assert_awaited_once_with(
            method="POST",
            path=path,
            body={"input": params["input"], "webhook_url": params["webhook_url"]},
            headers=headers,
        )

        if wait:
            mock_wait.assert_called_once_with(
                mock_prediction["id"], options=wait_options
            )
        else:
            mock_wait.assert_not_called()

        assert result == mock_api.async_request.return_value


@pytest.mark.parametrize(
    "status_sequence, expected_calls",
    [
        ([Status.pending, Status.started, Status.succeeded], 3),
        ([Status.succeeded], 1),
        ([Status.failed], 1),
    ],
)
@patch("time.sleep", return_value=None)
def test_wait(
    mock_sleep: Mock,
    status_sequence: list[Status],
    expected_calls: int,
) -> None:
    options = WaitOptions(poll_interval=2)
    base_prediction = PredictionWithIO(
        url="https://example.com",
        id="test_id",
        version_id="v1",
        created_at="2025-08-15T14:30:00Z",
        status=status_sequence[0],
        webhook_url=None,
        updated_at="2025-08-15T14:30:00Z",
        setup_time=None,
        execution_time=None,
        hardware_id="hw1",
        error=None,
        input={},
        output={},
        logs=None,
        setup_logs=None,
    )

    predictions = Predictions(api=Mock())

    responses = []
    for status in status_sequence:
        p = copy(base_prediction)
        p["status"] = status
        responses.append(p)

    with patch.object(predictions, "get", side_effect=responses) as mock_get:
        result = predictions.wait(base_prediction["id"], options=options)

    assert result["status"] == status_sequence[-1]
    assert mock_get.call_count == expected_calls
    assert mock_sleep.call_count == max(0, expected_calls - 1)
    mock_sleep.assert_has_calls(
        [call(options.poll_interval)] * max(0, expected_calls - 1)
    )


@pytest.mark.parametrize(
    "status_sequence, expected_calls",
    [
        ([Status.pending, Status.started, Status.succeeded], 3),
        ([Status.succeeded], 1),
        ([Status.failed], 1),
    ],
)
@patch("asyncio.sleep", new_callable=AsyncMock)
@pytest.mark.asyncio
async def test_async_wait(
    mock_sleep: AsyncMock,
    status_sequence: list[Status],
    expected_calls: int,
) -> None:
    options = WaitOptions(poll_interval=2)
    base_prediction = PredictionWithIO(
        url="https://example.com",
        id="test_id",
        version_id="v1",
        created_at="2025-08-15T14:30:00Z",
        status=status_sequence[0],
        webhook_url=None,
        updated_at="2025-08-15T14:30:00Z",
        setup_time=None,
        execution_time=None,
        hardware_id="hw1",
        error=None,
        input={},
        output={},
        logs=None,
        setup_logs=None,
    )

    predictions = Predictions(api=Mock())

    responses = []
    for status in status_sequence:
        p = copy(base_prediction)
        p["status"] = status
        responses.append(p)

    with patch.object(predictions, "async_get", side_effect=responses) as mock_get:
        result = await predictions.async_wait(base_prediction["id"], options=options)

    assert result["status"] == status_sequence[-1]
    assert mock_get.call_count == expected_calls
    assert mock_sleep.await_count == max(0, expected_calls - 1)
    mock_sleep.assert_has_awaits(
        [call(options.poll_interval)] * max(0, expected_calls - 1)
    )



from unittest.mock import Mock

from wriftai._resource import Resource


class MockAPIResource(Resource):
    """Concrete subclass for testing the abstract Resource."""

    pass


def test_api_resource() -> None:
    mock_api = Mock()
    resource = MockAPIResource(api=mock_api)
    assert resource._api == mock_api



"""Common types used across the WriftAI package."""

import sys
from collections.abc import Mapping
from enum import Enum
from typing import Any, TypeAlias, TypedDict, Union

__all__ = [
    "JsonValue",
    "NotRequired",
    "SchemaIO",
    "Schemas",
    "StrEnum",
    "User",
    "Version",
]

JsonValue: TypeAlias = Union[
    list["JsonValue"],
    Mapping[str, "JsonValue"],
    str,
    bool,
    int,
    float,
    None,
]
"""A JSON-compatible value."""


class _FallbackStrEnum(str, Enum):
    """Fallback StrEnum for python 3.10 version."""

    def __str__(self) -> str:
        """Return the enum value as string."""
        return str(self.value)


if sys.version_info >= (3, 11):
    from enum import StrEnum
    from typing import NotRequired
else:
    from enum import Enum

    from typing_extensions import NotRequired

    StrEnum = _FallbackStrEnum


class User(TypedDict):
    """Represents a user."""

    id: str
    """Unique identifier of the user."""
    username: str
    """The username of the user."""
    avatar_url: str
    """URL of the user's avatar."""
    name: str | None
    """The name of the user."""
    bio: str | None
    """The biography of the user."""
    urls: list[str] | None
    """Personal or professional website URLs."""
    location: str | None
    """Location of the user."""
    company: str | None
    """Company the user is associated with."""
    created_at: str
    """Timestamp when the user joined WriftAI."""
    updated_at: str | None
    """Timestamp when the user was last updated."""


class SchemaIO(TypedDict):
    """Represents input and output schemas."""

    input: dict[str, Any]
    """Schema for input, following JSON Schema Draft 2020-12 standards."""
    output: dict[str, Any]
    """Schema for output, following JSON Schema Draft 2020-12 standards."""


class Schemas(TypedDict):
    """Represents schemas of a version."""

    prediction: SchemaIO
    """The input and output schemas for a prediction."""


class Version(TypedDict):
    """Represents a version."""

    id: str
    """The unique identifier of the version."""
    release_notes: str
    """Information about changes such as new features,bug fixes,
        or optimizations in this version."""
    created_at: str
    """The time when the version was created."""
    schemas: Schemas
    """The schemas of the model version."""
    container_image_digest: str
    """A sha256 hash digest of the version's container image."""



from unittest.mock import AsyncMock, Mock, patch

import pytest

from wriftai.hardware import HardwareResource
from wriftai.pagination import PaginationOptions


@patch("wriftai.hardware.PaginatedResponse")
def test_list(mock_paginated_response: Mock) -> None:
    mock_api = Mock()
    test_response = {"key": "value"}
    mock_api.request.return_value = test_response

    hardware = HardwareResource(api=mock_api)
    pagination_options = PaginationOptions({"cursor": "abc123", "page_size": 50})
    result = hardware.list(pagination_options=pagination_options)

    mock_api.request.assert_called_once_with(
        method="GET", path=hardware._API_PREFIX, params=pagination_options
    )
    mock_paginated_response.assert_called_once_with(**test_response)
    assert result == mock_paginated_response.return_value


@patch("wriftai.hardware.PaginatedResponse")
@pytest.mark.asyncio
async def test_async_list(mock_paginated_response: Mock) -> None:
    mock_api = AsyncMock()
    test_response = {"key": "value"}
    mock_api.async_request.return_value = test_response

    hardware = HardwareResource(api=mock_api)
    pagination_options = PaginationOptions({"cursor": "abc123", "page_size": 50})
    result = await hardware.async_list(pagination_options=pagination_options)

    mock_api.async_request.assert_called_once_with(
        method="GET", path=hardware._API_PREFIX, params=pagination_options
    )
    mock_paginated_response.assert_called_once_with(**test_response)
    assert result == mock_paginated_response.return_value



import os
import re


def _extract_title_and_description(lines):
    """Extract title (heading) and description (first text paragraph) from Markdown."""
    title = None
    description = None

    for line in lines:
        stripped = line.strip()

        # Skip empty lines and YAML markers
        if not stripped or stripped == "---":
            continue

        # Markdown heading (e.g., "# abc module")
        if stripped.startswith("#"):
            raw_title = stripped.lstrip("#").strip()
            # Remove trailing "module" (case-insensitive)
            title = re.sub(
                r"\bmodule\b\.?$", "", raw_title, flags=re.IGNORECASE
            ).strip()
            continue

        # First non-empty non-heading line is description
        if not description:
            description = stripped

        if title and description:
            break

    # Fallbacks
    if not title and lines:
        raw_title = lines[0].strip()
        title = re.sub(r"\bmodule\b\.?$", "", raw_title, flags=re.IGNORECASE).strip()
    if not description:
        description = "No description available."

    return title, description


def add_frontmatter_to_docs() -> None:
    docs_path = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "docs", "reference")
    )

    if not os.path.exists(docs_path):
        print(f"Error: Documentation folder not found at {docs_path}")
        return

    for filename in os.listdir(docs_path):
        if not filename.endswith(".md"):
            continue

        filepath = os.path.join(docs_path, filename)

        with open(filepath, "r+", encoding="utf-8") as f:
            content = f.read().splitlines()

            # Skip files that already have YAML frontmatter
            if content and content[0].strip() == "---":
                continue

            # Custom handling for specific files
            if filename == "index.md":
                title = "Reference"
                description = "Reference for the client."
            elif filename == "wriftai.md":
                title = "Wriftai"
                description = "Package initializer for WriftAI Python Client."
            else:
                title, description = _extract_title_and_description(content)

            frontmatter = f"---\ntitle: {title}\ndescription: {description}\n---\n\n"

            new_content = frontmatter + "\n".join(content)
            f.seek(0)
            f.write(new_content)
            f.truncate()


if __name__ == "__main__":
    add_frontmatter_to_docs()



import enum as enum_module
import importlib
import sys
import typing as typing_module

import pytest

from wriftai.common_types import _FallbackStrEnum
from wriftai.users import UsersSortBy


@pytest.mark.parametrize("test_py_minor_version", [10, 11])
def test_StrEnum_import(
    monkeypatch: pytest.MonkeyPatch, test_py_minor_version: int
) -> None:
    # Simulate Python version to test conditional imports in common_types
    monkeypatch.setattr(sys, "version_info", (3, test_py_minor_version, 0))

    # Inject a mock StrEnum to allow import when simulating Python >= 3.11
    monkeypatch.setattr(enum_module, "StrEnum", enum_module.Enum, raising=False)

    # Stub typing.NotRequired if missing in simulated Python >= 3.11
    if test_py_minor_version >= 11 and not hasattr(typing_module, "NotRequired"):
        monkeypatch.setattr(typing_module, "NotRequired", object(), raising=False)

    import wriftai.common_types

    mod = importlib.reload(wriftai.common_types)

    assert hasattr(mod, "StrEnum")
    assert hasattr(mod, "NotRequired")
    assert str(UsersSortBy.CREATED_AT_ASC) == "created_at_asc"


def test_fallback_strenum() -> None:
    class TestEnum(_FallbackStrEnum):
        ENUM = "enum"

    assert str(TestEnum.ENUM) == "enum"



"""Client module."""

import os
from importlib.metadata import version
from typing import Any, Optional, TypedDict, TypeVar

import httpx

from wriftai.api import API
from wriftai.authenticated_user import AuthenticatedUser
from wriftai.hardware import HardwareResource
from wriftai.models import ModelsResource
from wriftai.predictions import Predictions
from wriftai.users import UsersResource
from wriftai.versions import Versions

TClient = TypeVar("TClient", httpx.Client, httpx.AsyncClient)

API_BASE_URL = "https://api.wrift.ai/v1"
AUTHORIZATION_HEADER = "Authorization"
USER_AGENT_HEADER = "User-Agent"


class ClientOptions(TypedDict):
    """Typed dictionary for specifying additional client options."""

    headers: dict[str, Any]
    """Optional HTTP headers to include in requests."""
    timeout: httpx.Timeout
    """Timeout configuration for requests.This should be an instance of
        `httpx.Timeout`."""
    transport: httpx.BaseTransport | None
    """Optional custom transport for managing HTTP behavior."""


class Client:
    """WriftAI client class.

    Attributes:
        api (API): The client used to interact with WriftAI's API.
        predictions (Predictions): Interface for accessing prediction related
            resources and operations.
        hardware (HardwareResource): Interface for hardware related resources
            and operations.
        authenticated_user (AuthenticatedUser): Interface for resources and operations
            related to the authenticated user.
        users (UsersResource): Interface for user related resources and
            operations.
        versions (Versions): Interface for resources and operations
            related to model versions.
        models (ModelsResource): Interface for resources and operations related
            to models.

    """

    def __init__(
        self,
        api_base_url: Optional[str] = None,
        access_token: Optional[str] = None,
        client_options: Optional[ClientOptions] = None,
    ) -> None:
        """Initializes a new instance of the Client class.

        Args:
            api_base_url (`Optional[str]`): The base URL for the API. If not provided,
                it falls back to the environment variable `WRIFTAI_API_BASE_URL` or
                the default api base url.
            access_token (`Optional[str]`): Bearer token for authorization. If not
                provided it falls back to the environment variable
                `WRIFTAI_API_ACCESS_TOKEN`.
            client_options (Optional[ClientOptions]): Additional options such as custom
                headers and timeout. Timeout defaults to 10s on all operations if not
                specified.
        """
        self._sync_client = _configure_client(
            client_type=httpx.Client,
            api_base_url=api_base_url,
            access_token=access_token,
            client_options=client_options,
        )
        self._async_client = _configure_client(
            client_type=httpx.AsyncClient,
            api_base_url=api_base_url,
            access_token=access_token,
            client_options=client_options,
        )
        self.api = API(sync_client=self._sync_client, async_client=self._async_client)
        self.predictions: Predictions = Predictions(api=self.api)
        self.hardware: HardwareResource = HardwareResource(api=self.api)
        self.authenticated_user: AuthenticatedUser = AuthenticatedUser(api=self.api)
        self.users: UsersResource = UsersResource(api=self.api)
        self.versions: Versions = Versions(api=self.api)
        self.models: ModelsResource = ModelsResource(api=self.api)


def _configure_client(
    client_type: type[TClient],
    api_base_url: str | None,
    access_token: str | None,
    client_options: ClientOptions | None,
) -> TClient:
    """Builds and returns a configured HTTPX client.

    This function sets up the HTTPX client with the specified base URL, headers,
    timeout, and authorization token. If api_base_url or access_token
    are not explicitly provided, it falls back to environment variables. The timeout
    is set to 10 seconds for all operations by default.

    Args:
        client_type (type[TClient]): The HTTPX client class to instantiate.
        api_base_url (str | None): The base URL for the API. If provided it overrides
            the value from the environment variable.
        access_token (str | None): Bearer token for authorization. If provided it
            overrides the value from the environment variable.
        client_options (ClientOptions | None): Additional client options like headers
            and timeout.

    Returns:
        httpx.Client | httpx.AsyncClient : A configured HTTPX client instance.
    """
    headers = (
        client_options["headers"]
        if client_options and "headers" in client_options
        else {}
    )
    timeout = (
        client_options["timeout"]
        if client_options and "timeout" in client_options
        else httpx.Timeout(10.0)
    )

    transport = (
        client_options["transport"]
        if client_options and "transport" in client_options
        else (
            httpx.HTTPTransport()
            if client_type is httpx.Client
            else httpx.AsyncHTTPTransport()
        )
    )

    access_token = access_token or os.environ.get("WRIFTAI_API_ACCESS_TOKEN")
    api_base_url = (
        api_base_url or os.environ.get("WRIFTAI_API_BASE_URL") or API_BASE_URL
    )

    if USER_AGENT_HEADER not in headers:
        headers[USER_AGENT_HEADER] = f"wriftai-python/{version('wriftai')}"

    if access_token and AUTHORIZATION_HEADER not in headers:
        headers[AUTHORIZATION_HEADER] = f"Bearer {access_token}"

    return client_type(
        base_url=api_base_url,
        headers=headers,
        timeout=timeout,
        transport=transport,  # type:ignore[arg-type]
    )



import json
from typing import Callable

import httpx
import pytest
from respx import Router

from tests.integration.conftest import Route
from wriftai._client import Client, ClientOptions
from wriftai.pagination import PaginatedResponse, PaginationOptions
from wriftai.predictions import CreatePredictionParams, Status


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_get(mock_router: Callable[..., Router], async_flag: bool) -> None:
    test_id = "c12258c4-ed83-4d7b-a784-8ed55412325a"
    expected_json = {
        "url": f"https://api.wriftai.com/v1/predictions/{test_id}",
        "id": test_id,
        "version_id": "abfb16e6-998d-49b9-941c-90dd14fc679f",
        "created_at": "2025-08-13T11:48:44.371093Z",
        "status": Status.pending,
        "webhook_url": None,
        "updated_at": "2025-08-13T11:48:44.371103Z",
        "setup_time": None,
        "execution_time": None,
        "hardware_id": "551c7d27-11e8-40a4-90d3-e8d8c22c5894",
        "error": None,
        "input": None,
        "output": None,
        "logs": None,
        "setup_logs": None,
    }

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1

    router = mock_router(
        route=Route(
            method="GET",
            path=f"/predictions/{test_id}",
            status_code=200,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        response = await client.predictions.async_get(prediction_id=test_id)
    else:
        response = client.predictions.get(prediction_id=test_id)

    assert dict(response) == expected_json
    assert call_count == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_list(mock_router: Callable[..., Router], async_flag: bool) -> None:
    expected_cursor = "abc123"
    expected_page_size = "10"
    expected_json = {
        "items": [
            {
                "url": "https://api.wriftai.com/v1/predictions",
                "id": "test_id",
                "version_id": "abfb16e6-998d-49b9-941c-90dd14fc679f",
                "created_at": "2025-08-13T11:48:44.371093Z",
                "status": Status.pending,
                "webhook_url": None,
                "updated_at": "2025-08-13T11:48:44.371103Z",
                "setup_time": None,
                "execution_time": None,
                "hardware_id": "551c7d27-11e8-40a4-90d3-e8d8c22c5894",
                "error": None,
            }
        ],
        "next_cursor": "abc123",
        "previous_cursor": None,
        "next_url": "/predictions?cursor=abc123",
        "previous_url": None,
    }

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1
        assert request.url.params.get("cursor") == expected_cursor
        assert request.url.params.get("page_size") == expected_page_size
        assert set(request.url.params.keys()).issubset({"cursor", "page_size"})

    router = mock_router(
        route=Route(
            method="GET",
            path="/predictions",
            status_code=200,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    pagination_options = PaginationOptions({
        "cursor": expected_cursor,
        "page_size": int(expected_page_size),
    })

    if async_flag:
        response = await client.predictions.async_list(
            pagination_options=pagination_options
        )
    else:
        response = client.predictions.list(pagination_options=pagination_options)

    # expected_json is a dictionary.
    # Although the structure matches the expected fields, static type checkers like mypy
    # may not be able to verify this due to dynamic typing.
    assert response == PaginatedResponse(**expected_json)  # type:ignore[arg-type]
    assert call_count == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_create_prediction_latest_version(
    mock_router: Callable[..., Router], async_flag: bool
) -> None:
    model_owner = "abc"
    model_name = "textgenerator"
    params: CreatePredictionParams = {
        "input": {"key": "value"},
        "webhook_url": "https://example.com/webhook",
        "validate_input": True,
    }

    expected_json = {
        "url": "https://api.wriftai.com",
        "id": "c12258c4-ed83-4d7b-a784-8ed55412325a",
        "version_id": "version-abc-123",
        "created_at": "2025-09-03T12:00:00Z",
        "status": Status.pending,
        "webhook_url": None,
        "updated_at": "2025-09-03T12:00:00Z",
        "setup_time": None,
        "execution_time": None,
        "hardware_id": "551c7d27-11e8-40a4-90d3-e8d8c22c5894",
        "error": None,
        "input": "test_input",
        "output": None,
        "logs": None,
        "setup_logs": None,
    }

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1
        json_payload = json.loads(request.content.decode())
        assert json_payload == params
        assert request.headers.get("validate-input") == "true"

    router = mock_router(
        route=Route(
            method="POST",
            path=f"/models/{model_owner}/{model_name}/predictions",
            status_code=202,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "Bearer test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        response = await client.predictions.async_create(
            model_owner=model_owner,
            model_name=model_name,
            params=params,
        )
    else:
        response = client.predictions.create(
            model_owner=model_owner,
            model_name=model_name,
            params=params,
        )

    assert response == expected_json
    assert call_count == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_create_prediction_specific_version(
    mock_router: Callable[..., Router], async_flag: bool
) -> None:
    version_id = "c12258c4-ed83-4d7b-a784-8ed55412325a"
    params: CreatePredictionParams = {
        "input": {"key": "value"},
    }

    expected_json = {
        "url": "https://api.wriftai.com",
        "id": "c12258c4-ed83-4d7b-a784-8ed55412325a",
        "version_id": version_id,
        "created_at": "2025-09-03T12:05:00Z",
        "status": Status.pending,
        "webhook_url": None,
        "updated_at": "2025-09-03T12:05:00Z",
        "setup_time": None,
        "execution_time": None,
        "hardware_id": "551c7d27-11e8-40a4-90d3-e8d8c22c5894",
        "error": None,
        "input": "test_input",
        "output": None,
        "logs": None,
        "setup_logs": None,
    }

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1

    router = mock_router(
        route=Route(
            method="POST",
            path=f"/versions/{version_id}/predictions",
            status_code=202,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "Bearer test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        response = await client.predictions.async_create(
            version_id=version_id,
            params=params,
        )
    else:
        response = client.predictions.create(
            version_id=version_id,
            params=params,
        )

    assert response == expected_json
    assert call_count == 1



from typing import Callable, Optional

import httpx
import pytest
from respx import Router

from tests.integration.conftest import Route
from wriftai._client import Client, ClientOptions
from wriftai.models import CreateModelParams, ModelVisibility, UpdateModelParams
from wriftai.pagination import PaginatedResponse


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
@pytest.mark.parametrize("owner", [None, "test_username"])
async def test_list(
    mock_router: Callable[..., Router], async_flag: bool, owner: Optional[str]
) -> None:
    expected_owner = owner or "test_username"
    path = f"/models/{expected_owner}" if owner else "/models"
    expected_json = {
        "items": [
            {
                "id": "123e4567-e89b-12d3-a456-426614174000",
                "name": "test_model_name",
                "created_at": "2025-08-01T10:00:00Z",
                "visibility": ModelVisibility.public,
                "description": "A computer vision model.",
                "updated_at": "2025-08-15T14:30:00Z",
                "source_url": "https://example.com/source",
                "license_url": "https://example.com/license",
                "paper_url": "https://example.com/paper",
                "owner": {
                    "id": "user-001",
                    "username": expected_owner,
                    "avatar_url": "https://example.com/avatar.png",
                    "name": "dummy user",
                    "bio": "Soft Dev",
                    "urls": ["https://github.com/dummy"],
                    "location": "Karachi",
                    "company": "Sych",
                    "created_at": "2024-05-10T09:00:00Z",
                    "updated_at": "2025-07-20T16:45:00Z",
                },
                "latest_version": {
                    "id": "ver-001",
                    "release_notes": "Initial release.",
                    "created_at": "2025-08-10T09:00:00Z",
                    "schemas": {
                        "prediction": {
                            "input": {"key1": "value1", "key2": 123},
                            "output": {"result": True, "message": "Success"},
                        }
                    },
                    "container_image_digest": "sha256:abc123def456ghi",
                },
                "hardware_name": "NVIDIA A100",
                "predictions_count": 2048,
            }
        ],
        "next_cursor": "abc123",
        "previous_cursor": None,
        "next_url": "/models?cursor=abc123",
        "previous_url": None,
    }

    router = mock_router(
        route=Route(
            method="GET",
            path=path,
            status_code=200,
            json=expected_json,
        )
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        response = await client.models.async_list(owner=owner)
    else:
        response = client.models.list(owner=owner)

    # expected_json is a dictionary.
    # Although the structure matches the expected fields, static type checkers like mypy
    # may not be able to verify this due to dynamic typing.
    assert response == PaginatedResponse(**expected_json)  # type:ignore[arg-type]


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_delete(mock_router: Callable[..., Router], async_flag: bool) -> None:
    test_owner = "test_owner"
    test_model_name = "dummy_model"

    router = mock_router(
        route=Route(
            method="DELETE",
            path=f"/models/{test_owner}/{test_model_name}",
            status_code=204,
            json={},
        )
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        await client.models.async_delete(owner=test_owner, name=test_model_name)
    else:
        client.models.delete(owner=test_owner, name=test_model_name)


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_get(mock_router: Callable[..., Router], async_flag: bool) -> None:
    test_owner = "test_owner"
    test_model_name = "dummy_model"

    expected_json = {
        "id": "123e4567-e89b-12d3-a456-426614174000",
        "name": test_model_name,
        "created_at": "2025-08-01T10:00:00Z",
        "visibility": ModelVisibility.public,
        "description": "A computer vision model.",
        "updated_at": "2025-08-15T14:30:00Z",
        "source_url": "https://example.com/source",
        "license_url": "https://example.com/license",
        "paper_url": "https://example.com/paper",
        "owner": {
            "id": "user-001",
            "username": test_owner,
            "avatar_url": "https://example.com/avatar.png",
            "name": "dummy user",
            "bio": "Soft Dev",
            "urls": ["https://github.com/dummy"],
            "location": "Karachi",
            "company": "Sych",
            "created_at": "2024-05-10T09:00:00Z",
            "updated_at": "2025-07-20T16:45:00Z",
        },
        "latest_version": {
            "id": "ver-001",
            "release_notes": "Initial release.",
            "created_at": "2025-08-10T09:00:00Z",
            "schemas": {
                "prediction": {
                    "input": {"key1": "value1", "key2": 123},
                    "output": {"result": True, "message": "Success"},
                }
            },
            "container_image_digest": "sha256:abc123def456ghi",
        },
        "hardware_name": "NVIDIA A100",
        "predictions_count": 2048,
    }

    router = mock_router(
        route=Route(
            method="GET",
            path=f"/models/{test_owner}/{test_model_name}",
            status_code=200,
            json=expected_json,
        )
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        response = await client.models.async_get(owner=test_owner, name=test_model_name)
    else:
        response = client.models.get(owner=test_owner, name=test_model_name)

    assert response == expected_json


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_create(mock_router: Callable[..., Router], async_flag: bool) -> None:
    params: CreateModelParams = {
        "name": "test_model",
        "visibility": ModelVisibility.private,
        "hardware_name": "NVIDIA A100",
        "description": "Test model description",
        "source_url": "https://example.com/source",
        "license_url": "https://example.com/license",
        "paper_url": "https://example.com/paper",
    }

    expected_json = {
        "id": "123e4567-e89b-12d3-a456-426614174000",
        "name": params["name"],
        "created_at": "2025-08-01T10:00:00Z",
        "visibility": params["visibility"],
        "description": params["description"],
        "updated_at": "2025-08-15T14:30:00Z",
        "source_url": params["source_url"],
        "license_url": params["license_url"],
        "paper_url": params["paper_url"],
        "owner": {
            "id": "user-001",
            "username": "Kunal_Kumar",
            "avatar_url": "https://example.com/avatar.png",
            "name": "Kunal Kumar",
            "bio": "Soft Dev",
            "urls": ["https://github.com/kunal"],
            "location": "Karachi",
            "company": "Sych",
            "created_at": "2024-05-10T09:00:00Z",
            "updated_at": "2025-07-20T16:45:00Z",
        },
        "latest_version": {
            "id": "ver-001",
            "release_notes": "Initial release.",
            "created_at": "2025-08-10T09:00:00Z",
            "schemas": {
                "prediction": {
                    "input": {"key1": "value1", "key2": 123},
                    "output": {"result": True, "message": "Success"},
                }
            },
            "container_image_digest": "sha256:abc123def456ghi",
        },
        "hardware_name": params["hardware_name"],
        "predictions_count": 2048,
    }

    router = mock_router(
        route=Route(
            method="POST",
            path="/models",
            status_code=200,
            json=expected_json,
        )
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        response = await client.models.async_create(params)
    else:
        response = client.models.create(params)
    assert response == expected_json


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_update(mock_router: Callable[..., Router], async_flag: bool) -> None:
    test_owner = "test_owner"
    test_model_name = "dummy_model"
    description = "Updated description"
    source_url = "https://example.com/updated_source"
    license_url = "https://license.com/updated_license"
    paper_url = "https://paper.com/updated_paper"
    hardware_name = "Updated Hardware"

    expected_json = {
        "id": "123e4567-e89b-12d3-a456-426614174000",
        "name": test_model_name,
        "created_at": "2025-08-01T10:00:00Z",
        "visibility": ModelVisibility.public,
        "description": description,
        "updated_at": "2025-08-15T14:30:00Z",
        "source_url": source_url,
        "license_url": license_url,
        "paper_url": paper_url,
        "owner": {
            "id": "user-001",
            "username": test_owner,
            "avatar_url": "https://example.com/avatar.png",
            "name": "dummy user",
            "bio": "Soft Dev",
            "urls": ["https://github.com/kunal"],
            "location": "Karachi",
            "company": "Sych",
            "created_at": "2024-05-10T09:00:00Z",
            "updated_at": "2025-07-20T16:45:00Z",
        },
        "latest_version": {
            "id": "ver-001",
            "release_notes": "Initial release.",
            "created_at": "2025-08-10T09:00:00Z",
            "schemas": {
                "prediction": {
                    "input": {"key1": "value1", "key2": 123},
                    "output": {"result": True, "message": "Success"},
                }
            },
            "container_image_digest": "sha256:abc123def456ghi",
        },
        "hardware_name": hardware_name,
        "predictions_count": 2048,
    }
    payload: UpdateModelParams = {
        "description": description,
        "source_url": source_url,
        "license_url": license_url,
        "paper_url": paper_url,
        "hardware_name": hardware_name,
    }
    router = mock_router(
        route=Route(
            method="PATCH",
            path=f"/models/{test_owner}/{test_model_name}",
            status_code=200,
            json=expected_json,
        )
    )
    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )
    if async_flag:
        response = await client.models.async_update(
            owner=test_owner,
            name=test_model_name,
            params=payload,
        )
    else:
        response = client.models.update(
            owner=test_owner,
            name=test_model_name,
            params=payload,
        )
    assert response == expected_json


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_search(mock_router: Callable[..., Router], async_flag: bool) -> None:
    expected_json = {
        "items": [
            {
                "id": "123e4567-e89b-12d3-a456-426614174000",
                "name": "test_model_name",
                "created_at": "2025-08-01T10:00:00Z",
                "visibility": ModelVisibility.public,
                "description": "A computer vision model.",
                "updated_at": "2025-08-15T14:30:00Z",
                "source_url": "https://example.com/source",
                "license_url": "https://example.com/license",
                "paper_url": "https://example.com/paper",
                "owner": {
                    "id": "user-001",
                    "username": "test_username",
                    "avatar_url": "https://example.com/avatar.png",
                    "name": "dummy user",
                    "bio": "Soft Dev",
                    "urls": ["https://github.com/dummy"],
                    "location": "Karachi",
                    "company": "Sych",
                    "created_at": "2024-05-10T09:00:00Z",
                    "updated_at": "2025-07-20T16:45:00Z",
                },
                "latest_version": {
                    "id": "ver-001",
                    "release_notes": "Initial release.",
                    "created_at": "2025-08-10T09:00:00Z",
                    "schemas": {
                        "prediction": {
                            "input": {"key1": "value1", "key2": 123},
                            "output": {"result": True, "message": "Success"},
                        }
                    },
                    "container_image_digest": "sha256:abc123def456ghi",
                },
                "hardware_name": "NVIDIA A100",
                "predictions_count": 2048,
            }
        ],
        "next_cursor": "abc123",
        "previous_cursor": None,
        "next_url": "/search/models?cursor=abc123",
        "previous_url": None,
    }

    router = mock_router(
        route=Route(
            method="GET",
            path="/search/models",
            status_code=200,
            json=expected_json,
        )
    )
    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        response = await client.models.async_search(q="dummy_query")
    else:
        response = client.models.search(q="dummy_query")

    # expected_json is a dictionary.
    # Although the structure matches the expected fields, static type checkers like mypy
    # may not be able to verify this due to dynamic typing.
    assert response == PaginatedResponse(**expected_json)  # type:ignore[arg-type]



from unittest.mock import Mock, call, patch

import httpx

from wriftai._client import (
    API_BASE_URL,
    AUTHORIZATION_HEADER,
    USER_AGENT_HEADER,
    Client,
    ClientOptions,
    _configure_client,
)


@patch("wriftai._client._configure_client")
@patch("wriftai._client.Predictions")
@patch("wriftai._client.API")
@patch("wriftai._client.AuthenticatedUser")
@patch("wriftai._client.UsersResource")
@patch("wriftai._client.HardwareResource")
@patch("wriftai._client.ModelsResource")
@patch("wriftai._client.Versions")
def test_client_class(
    mock_versions: Mock,
    mock_models_resource: Mock,
    mock_hardware: Mock,
    mock_users_resource: Mock,
    mock_authenticated_user: Mock,
    mock_api: Mock,
    mock_predictions: Mock,
    mock__configure_client: Mock,
) -> None:
    test_api_base_url = "https://api.wrift.ai"
    test_access_token = "A9f3D4eH2jK7LmN0Q1rTzX8vY6bC5uW4"  # noqa:S105
    test_client_options = ClientOptions(
        headers={AUTHORIZATION_HEADER: "test-auth"},
        timeout=httpx.Timeout(20),
        transport=Mock(),
    )
    mock__sync_client = Mock()
    mock__async_client = Mock()

    mock__configure_client.side_effect = [mock__sync_client, mock__async_client]

    mock_client = Client(
        api_base_url=test_api_base_url,
        access_token=test_access_token,
        client_options=test_client_options,
    )

    assert mock_client._sync_client == mock__sync_client
    assert mock_client._async_client == mock__async_client
    assert mock_client.api == mock_api.return_value
    assert mock_client.predictions == mock_predictions.return_value
    assert mock_client.hardware == mock_hardware.return_value
    assert mock_client.versions == mock_versions.return_value
    assert mock_client.authenticated_user == mock_authenticated_user.return_value
    assert mock_client.users == mock_users_resource.return_value
    assert mock_client.models == mock_models_resource.return_value

    mock__configure_client.assert_has_calls([
        call(
            client_type=httpx.Client,
            api_base_url=test_api_base_url,
            access_token=test_access_token,
            client_options=test_client_options,
        ),
        call(
            client_type=httpx.AsyncClient,
            api_base_url=test_api_base_url,
            access_token=test_access_token,
            client_options=test_client_options,
        ),
    ])
    mock_api.assert_called_once_with(
        sync_client=mock__sync_client, async_client=mock__async_client
    )
    mock_predictions.assert_called_once_with(api=mock_api.return_value)
    mock_users_resource.assert_called_once_with(api=mock_api.return_value)
    mock_hardware.assert_called_once_with(api=mock_api.return_value)
    mock_versions.assert_called_once_with(api=mock_api.return_value)
    mock_authenticated_user.assert_called_once_with(api=mock_api.return_value)
    mock_models_resource.assert_called_once_with(api=mock_api.return_value)


@patch("wriftai._client.os")
@patch("wriftai._client.version")
@patch("wriftai._client.httpx.HTTPTransport")
@patch("wriftai._client.httpx.AsyncHTTPTransport")
def test__configure_client_with_None_args(
    mock_async_http_transport: Mock,
    mock_http_transport: Mock,
    mock_version: Mock,
    mock_os: Mock,
) -> None:
    mock_client_type = Mock()
    test_env_var = "test-env-var"

    mock_os.environ.get.return_value = test_env_var

    result = _configure_client(
        client_type=mock_client_type,
        api_base_url=None,
        access_token=None,
        client_options=None,
    )

    mock_os.environ.get.assert_has_calls([
        call("WRIFTAI_API_ACCESS_TOKEN"),
        call("WRIFTAI_API_BASE_URL"),
    ])

    mock_client_type.assert_called_once_with(
        base_url=test_env_var,
        headers={
            USER_AGENT_HEADER: f"wriftai-python/{mock_version.return_value}",
            AUTHORIZATION_HEADER: f"Bearer {test_env_var}",
        },
        timeout=httpx.Timeout(10.0),
        transport=mock_async_http_transport.return_value,
    )
    mock_http_transport.assert_not_called()
    assert result == mock_client_type.return_value


@patch("wriftai._client.os")
def test__configure_client_with_args_provided(mock_os: Mock) -> None:
    mock_client_type = Mock()
    test_api_base_url = "https://api.wrift.ai"
    test_access_token = "A9f3D4eH2jK7LmN0Q1rTzX8vY6bC5uW4"  # noqa:S105
    mock_transport = Mock()

    test_client_options = ClientOptions(
        headers={
            USER_AGENT_HEADER: "wriftai-python/0.0.1",
            AUTHORIZATION_HEADER: f"Bearer {test_access_token}",
        },
        timeout=httpx.Timeout(20.0),
        transport=mock_transport,
    )

    result = _configure_client(
        client_type=mock_client_type,
        api_base_url=test_api_base_url,
        access_token=test_access_token,
        client_options=test_client_options,
    )

    mock_os.environ.get.assert_not_called()

    mock_client_type.assert_called_once_with(
        base_url=test_api_base_url,
        headers=test_client_options["headers"],
        timeout=test_client_options["timeout"],
        transport=mock_transport,
    )
    assert result == mock_client_type.return_value


@patch("wriftai._client.os")
@patch("wriftai._client.version")
def test__configure_client_with_args_and_env_vars_provided(
    mock_version: Mock, mock_os: Mock
) -> None:
    mock_client_type = Mock()
    test_api_base_url = "https://api.wrift.ai"
    test_access_token = "A9f3D4eH2jK7LmN0Q1rTzX8vY6bC5uW4"  # noqa:S105
    mock_transport = Mock()

    test_client_options = ClientOptions(
        headers={
            USER_AGENT_HEADER: "wriftai-python/0.0.1",
            AUTHORIZATION_HEADER: f"Bearer {test_access_token}",
        },
        timeout=httpx.Timeout(20.0),
        transport=mock_transport,
    )

    mock_os.environ.get.return_value = "test-env-var"

    result = _configure_client(
        client_type=mock_client_type,
        api_base_url=test_api_base_url,
        access_token=test_access_token,
        client_options=test_client_options,
    )

    # assert environment variables not used.
    mock_os.environ.get.assert_not_called()
    # assert version not called.
    mock_version.assert_not_called()

    mock_client_type.assert_called_once_with(
        base_url=test_api_base_url,
        headers=test_client_options["headers"],
        timeout=test_client_options["timeout"],
        transport=mock_transport,
    )
    assert result == mock_client_type.return_value


@patch("wriftai._client.os")
def test__configure_client_with_no_args_and_no_env_var(
    mock_os: Mock,
) -> None:
    mock_client_type = Mock()
    mock_transport = Mock()

    test_client_options = ClientOptions(
        headers={
            USER_AGENT_HEADER: "wriftai-python/0.0.1",
        },
        timeout=httpx.Timeout(20.0),
        transport=mock_transport,
    )

    # no env variable set.
    mock_os.environ.get.return_value = None

    result = _configure_client(
        client_type=mock_client_type,
        api_base_url=None,
        access_token=None,
        client_options=test_client_options,
    )

    mock_os.environ.get.assert_has_calls([
        call("WRIFTAI_API_ACCESS_TOKEN"),
        call("WRIFTAI_API_BASE_URL"),
    ])

    mock_client_type.assert_called_once_with(
        base_url=API_BASE_URL,
        headers=test_client_options["headers"],
        timeout=test_client_options["timeout"],
        transport=mock_transport,
    )
    assert result == mock_client_type.return_value



import json
import re
from typing import Callable

import httpx
import pytest
from respx import Router

from tests.integration.conftest import Route
from wriftai._client import Client, ClientOptions


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_request(mock_router: Callable[..., Router], async_flag: bool) -> None:
    expected_json = {"key": "value"}
    expected_method = "POST"
    expected_path = "/request"
    test_headers = {"headers-key": "headers-value"}
    test_body = {"body-key": "body-value"}
    test_user_agent = r"^wriftai-python/\d+\.\d+\.\d+(-[\w.-]+)?$"
    expected_access_token = "some-access-token"  # noqa: S105

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1
        assert request.method == expected_method
        assert request.url.path == f"/v1{expected_path}"
        assert re.match(test_user_agent, request.headers.get("user-agent"))
        assert request.headers.get("content-type") == "application/json"
        assert request.headers.get("headers-key") == "headers-value"
        assert request.headers.get("Authorization") == f"Bearer {expected_access_token}"
        json_payload = json.loads(request.content.decode())
        assert json_payload == test_body

    router = mock_router(
        route=Route(
            method="POST",
            path=expected_path,
            status_code=200,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        access_token=expected_access_token,
        client_options=ClientOptions(
            headers={},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        ),
    )

    if async_flag:
        response = await client.api.async_request(
            method=expected_method,
            path=expected_path,
            body=test_body,
            headers=test_headers,
        )
    else:
        response = client.api.request(
            method=expected_method,
            path=expected_path,
            body=test_body,
            headers=test_headers,
        )

    assert response == expected_json
    assert call_count == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_request_raises_404_error(
    mock_router: Callable[..., Router], async_flag: bool
) -> None:
    test_method = "POST"
    test_path = "/request"
    test_error_body = {"error": "Request not successful."}
    test_status = 404
    router = mock_router(
        route=Route(
            method=test_method,
            path=test_path,
            status_code=test_status,
            json=test_error_body,
        )
    )

    client = Client(
        client_options=ClientOptions(
            headers={},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    with pytest.raises(httpx.HTTPStatusError) as e:
        if async_flag:
            await client.api.async_request(method=test_method, path=test_path)
        else:
            client.api.request(method=test_method, path=test_path)

    assert (
        str(e.value)
        == "Client error '404 Not Found' for url "
        + f"'https://api.wrift.ai/v1{test_path}'\n"
        + "For more information check: https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/404"
    )
    assert e.value.response.status_code == test_status
    assert e.value.response.json() == test_error_body



from unittest.mock import AsyncMock, Mock, patch

import pytest

from wriftai.authenticated_user import AuthenticatedUser, UpdateUserParams
from wriftai.pagination import PaginationOptions


def test_get() -> None:
    mock_api = Mock()
    mock_api.request = Mock()

    user = AuthenticatedUser(api=mock_api)

    result = user.get()

    mock_api.request.assert_called_once_with("GET", f"{user._API_PREFIX}")
    assert result == mock_api.request.return_value


@pytest.mark.asyncio
async def test_async_get() -> None:
    mock_api = Mock()
    mock_api.async_request = AsyncMock()

    user = AuthenticatedUser(api=mock_api)

    result = await user.async_get()

    mock_api.async_request.assert_called_once_with("GET", f"{user._API_PREFIX}")
    assert result == mock_api.async_request.return_value


@patch("wriftai.authenticated_user.PaginatedResponse")
def test_models(mock_paginated_response: Mock) -> None:
    mock_api = Mock()
    test_response = {"key": "value"}
    mock_api.request.return_value = test_response

    user = AuthenticatedUser(api=mock_api)
    pagination_options = PaginationOptions({"cursor": "abc123", "page_size": 50})
    result = user.models(pagination_options=pagination_options)

    mock_api.request.assert_called_once_with(
        method="GET",
        path=f"{user._API_PREFIX}{user._MODELS_API_PREFIX}",
        params=pagination_options,
    )
    mock_paginated_response.assert_called_once_with(**test_response)
    assert result == mock_paginated_response.return_value


@pytest.mark.asyncio
@patch("wriftai.authenticated_user.PaginatedResponse")
async def test_async_models(mock_paginated_response: Mock) -> None:
    mock_api = AsyncMock()
    test_response = {"key": "value"}
    mock_api.async_request.return_value = test_response

    user = AuthenticatedUser(api=mock_api)
    pagination_options = PaginationOptions({"cursor": "abc123", "page_size": 50})
    result = await user.async_models(pagination_options=pagination_options)

    mock_api.async_request.assert_called_once_with(
        method="GET",
        path=f"{user._API_PREFIX}{user._MODELS_API_PREFIX}",
        params=pagination_options,
    )
    mock_paginated_response.assert_called_once_with(**test_response)
    assert result == mock_paginated_response.return_value


def test_update() -> None:
    mock_api = Mock()
    user = AuthenticatedUser(api=mock_api)
    payload: UpdateUserParams = {
        "username": "test_username",
        "name": "dummy_user",
        "company": "Dummy Company",
        "location": "Nowhere",
        "bio": "This is a dummy user.",
        "urls": ["https://example.com", "https://example.org"],
    }

    result = user.update(
        params=payload,
    )
    mock_api.request.assert_called_once_with(
        method="PATCH",
        path=f"{user._API_PREFIX}",
        body=payload,
    )
    assert result == mock_api.request.return_value


@pytest.mark.asyncio
async def test_async_update() -> None:
    mock_api = AsyncMock()
    user = AuthenticatedUser(api=mock_api)
    payload: UpdateUserParams = {
        "name": "dummy_user",
        "company": "Dummy Company",
        "location": "Nowhere",
    }
    result = await user.async_update(
        params=payload,
    )
    mock_api.async_request.assert_called_once_with(
        method="PATCH",
        path=f"{user._API_PREFIX}",
        body=payload,
    )
    assert result == mock_api.async_request.return_value



"""conf.py."""

import os
import sys
from enum import EnumMeta, StrEnum
from typing import Any

from sphinx.application import Sphinx

sys.path.insert(0, os.path.abspath("../wriftai"))

# Sphinx extensions to enable
extensions = [
    "sphinx.ext.autodoc",  # Automatically generate docs from docstrings
    "sphinx.ext.napoleon",  # Support Google- and NumPy-style docstrings
    "sphinx_markdown_builder",  # Enable Markdown output instead of HTML
]

# Napoleon settings for docstring parsing
napoleon_google_docstring = True  # Parse Google-style docstrings
napoleon_use_ivar = (
    True  # Treat class docstring attributes as instance variables in docs
)

# Autodoc settings
autodoc_typehints = "description"  # Render type hints in the description section

autodoc_member_order = (
    "bysource"  # Document members in the order they appear in the source code
)
autodoc_show_inheritance = True  # Show base classes for inherited classes
# General formatting
add_module_names = (
    False  # Don't prepend module names to class/function references in docs
)

autoclass_content = "init"  # Render class description from init

# Custom type aliases for autodoc
autodoc_type_aliases = {"JsonValue": "JsonValue"}
markdown_anchor_signatures = True


def shorten_titles(app: Sphinx, *args: Any) -> None:
    """Remove wriftai. from ALL titles in docs/."""
    srcdir = app.srcdir

    for filename in os.listdir(srcdir):
        if not filename.endswith(".rst"):
            continue

        path = os.path.join(srcdir, filename)

        with open(path) as f:
            lines = f.readlines()

        if not lines or "wriftai." not in lines[0]:
            continue

        short = lines[0].strip().replace("wriftai.", "")
        lines[0] = short + "\n"
        lines[1] = "=" * len(short) + "\n"

        with open(path, "w") as f:
            f.writelines(lines)


def fix_enum_signatures(app, what, name, obj, options, signature, return_annotation):
    """Fixes the signatures of StrEnum and Enum classes in Sphinx autodoc.

    Sphinx normally expands StrEnum/Enum classes into their internal `EnumMeta`
    constructor signature, showing parameters like
    `(new_class_name, names, module, ...)`. This clutters the documentation
    for end users. This hook replaces the noisy internal signature with a
    clean, user-friendly one.
    """
    if isinstance(obj, EnumMeta):
        if obj is StrEnum:
            return "(str, Enum)", return_annotation
        if issubclass(obj, StrEnum):
            return "(StrEnum)", return_annotation
        return "(Enum)", return_annotation
    return signature, return_annotation


def setup(app: Sphinx) -> None:
    """Connect custom Sphinx hooks to modify documentation during the build."""
    app.connect("env-before-read-docs", shorten_titles)
    app.connect("autodoc-process-signature", fix_enum_signatures)



"""Predictions module."""

import asyncio
import time
from dataclasses import dataclass
from typing import Optional, TypedDict, cast, overload

from wriftai._resource import Resource
from wriftai.common_types import JsonValue, NotRequired, StrEnum
from wriftai.pagination import PaginatedResponse, PaginationOptions


class ErrorSource(StrEnum):
    """Enumeration of possible error sources."""

    internal = "internal"
    external = "external"


class Status(StrEnum):
    """Enumeration of possible prediction statuses."""

    pending = "pending"
    started = "started"
    failed = "failed"
    succeeded = "succeeded"


class TaskError(TypedDict):
    """Represents an error that occurred during a prediction."""

    source: ErrorSource
    message: str
    detail: JsonValue


class Prediction(TypedDict):
    """Represents a prediction."""

    url: str
    """URL to access the prediction resource."""
    id: str
    """Unique identifier for the prediction."""
    version_id: str
    """Identifier of the model version used."""
    created_at: str
    """Timestamp when the prediction was created."""
    status: Status
    """Current status of the prediction."""
    webhook_url: str | None
    """URL for webhook callbacks, if any."""
    updated_at: str
    """Timestamp of the last update to the prediction."""
    setup_time: str | None
    """Time taken to set up the prediction environment."""
    execution_time: str | None
    """Time taken to execute the prediction."""
    hardware_id: str
    """Identifier of the hardware used for prediction."""
    error: TaskError | None
    """Error details if the prediction failed."""


class PredictionWithIO(Prediction):
    """Represents a prediction with I/O details."""

    input: JsonValue
    """Input data provided for the prediction."""
    output: JsonValue
    """Output generated by the prediction."""
    logs: list[str] | None
    """Logs generated during prediction execution."""
    setup_logs: list[str] | None
    """Logs generated during setup."""


class CreatePredictionParams(TypedDict):
    """Prediction creation params."""

    input: JsonValue
    """Input data provided for the prediction."""
    webhook_url: NotRequired[str]
    """An HTTP URL to POST prediction updates to."""
    validate_input: NotRequired[bool]
    """Enable input validation against the schema before processing."""


@dataclass
class WaitOptions:
    """Options for customizing wait behaviour."""

    poll_interval: float = 1.0
    """Time in seconds between polling attempts to check the prediction status."""


DEFAULT_WAIT_OPTIONS = WaitOptions()
"""Default wait options."""


class Predictions(Resource):
    """Resource for operations related to predictions."""

    _API_PREFIX = "/predictions"
    _PREDICTIONS_API_SUFFIX = "/predictions"
    _ERROR_MSG_INVALID_PREDICTION_PARAMS = (
        "Either a version_id or both model_owner and model_name must be provided."
    )

    def get(self, prediction_id: str) -> PredictionWithIO:
        """Get a prediction by id.

        Args:
            prediction_id (str): The unique identifier of the prediction.

        Returns:
            PredictionWithIO: The prediction object.
        """
        return cast(
            PredictionWithIO,
            self._api.request(method="GET", path=f"{self._API_PREFIX}/{prediction_id}"),
        )

    async def async_get(self, prediction_id: str) -> PredictionWithIO:
        """Get a prediction by id.

        Args:
            prediction_id (str): The unique identifier of the prediction.

        Returns:
            PredictionWithIO: The prediction object.
        """
        return cast(
            PredictionWithIO,
            await self._api.async_request(
                method="GET", path=f"{self._API_PREFIX}/{prediction_id}"
            ),
        )

    def list(
        self,
        pagination_options: Optional[PaginationOptions] = None,
    ) -> PaginatedResponse[Prediction]:
        """List predictions.

        Args:
            pagination_options (Optional[PaginationOptions]): Optional settings
                to control pagination behavior.

        Returns:
            PaginatedResponse[Prediction]: Paginated response containing predictions
                and navigation metadata.
        """
        response = self._api.request(
            method="GET", params=pagination_options, path=self._API_PREFIX
        )

        # The response will always match the PaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]

    async def async_list(
        self,
        pagination_options: Optional[PaginationOptions] = None,
    ) -> PaginatedResponse[Prediction]:
        """List predictions.

        Args:
            pagination_options (Optional[PaginationOptions]): Optional settings
                to control pagination behavior.

        Returns:
            PaginatedResponse[Prediction]: Paginated response containing predictions
                and navigation metadata.
        """
        response = await self._api.async_request(
            method="GET", params=pagination_options, path=self._API_PREFIX
        )
        # The response will always match the PaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]

    def _prediction_path(
        self,
        model_owner: Optional[str] = None,
        model_name: Optional[str] = None,
        version_id: Optional[str] = None,
    ) -> str:
        """Constructs the API path to create a prediction.

        Args:
            model_owner (Optional[str]): Username of the model's owner.
            model_name (Optional[str]): Name of the model.
            version_id (Optional[str]): Unique identifier of the model's version.

        Returns:
            str: The API path.

        Raises:
            TypeError: When an invalid combination of version_id, model_owner,
                and model_name is provided.
        """
        if version_id is None and model_owner is not None and model_name is not None:
            return (
                f"{self._MODELS_API_PREFIX}/{model_owner}/"
                f"{model_name}{self._PREDICTIONS_API_SUFFIX}"
            )
        elif model_owner is None and model_name is None and version_id is not None:
            return (
                f"{self._VERSIONS_API_PREFIX}/{version_id}"
                f"{self._PREDICTIONS_API_SUFFIX}"
            )
        else:
            raise TypeError(self._ERROR_MSG_INVALID_PREDICTION_PARAMS)

    @overload
    def create(
        self,
        *,
        model_owner: str,
        model_name: str,
        wait: bool = False,
        wait_options: WaitOptions = DEFAULT_WAIT_OPTIONS,
        params: CreatePredictionParams,
    ) -> PredictionWithIO: ...

    @overload
    def create(
        self,
        *,
        version_id: str,
        wait: bool = False,
        wait_options: WaitOptions = DEFAULT_WAIT_OPTIONS,
        params: CreatePredictionParams,
    ) -> PredictionWithIO: ...

    def create(
        self,
        params: CreatePredictionParams,
        model_owner: Optional[str] = None,
        model_name: Optional[str] = None,
        version_id: Optional[str] = None,
        wait: bool = False,
        wait_options: WaitOptions = DEFAULT_WAIT_OPTIONS,
    ) -> PredictionWithIO:
        """Create a prediction.

        This method creates a prediction using either the latest version of a model
        or a specific model version.

        Args:
            model_owner (`Optional[str]`): Username of the model's owner.
            model_name (`Optional[str]`): Name of the model.
            version_id (`Optional[str]`): Unique identifier of the model's version.
            params (CreatePredictionParams): Prediction creation params.
            wait (bool): If True, waits until the prediction reaches a terminal state.
                If False, returns the prediction immediately. Defaults to False.
            wait_options (WaitOptions): Options for customizing wait behavior.
                If not provided, default options are used.

        Returns:
            PredictionWithIO: The new prediction.

        Raises:
            TypeError: When an invalid combination of version_id, model_owner,
                and model_name is provided.
        """
        path = self._prediction_path(model_owner, model_name, version_id)

        headers = None
        if "validate_input" in params:
            headers = {"Validate-Input": str(params.pop("validate_input")).lower()}

        prediction = cast(
            PredictionWithIO,
            self._api.request(
                method="POST",
                path=path,
                # The params matches JsonValue at runtime,
                # but static type checkers may not infer this correctly.
                # Hence, we ignore the argument type warning.
                body=params,  # type:ignore[arg-type]
                headers=headers,
            ),
        )

        if wait:
            return self.wait(prediction["id"], options=wait_options)

        return prediction

    @overload
    async def async_create(
        self,
        *,
        model_owner: str,
        model_name: str,
        wait: bool = False,
        wait_options: WaitOptions = DEFAULT_WAIT_OPTIONS,
        params: CreatePredictionParams,
    ) -> PredictionWithIO: ...

    @overload
    async def async_create(
        self,
        *,
        version_id: str,
        wait: bool = False,
        wait_options: WaitOptions = DEFAULT_WAIT_OPTIONS,
        params: CreatePredictionParams,
    ) -> PredictionWithIO: ...

    async def async_create(
        self,
        params: CreatePredictionParams,
        model_owner: Optional[str] = None,
        model_name: Optional[str] = None,
        version_id: Optional[str] = None,
        wait: bool = False,
        wait_options: WaitOptions = DEFAULT_WAIT_OPTIONS,
    ) -> PredictionWithIO:
        """Create a prediction.

        This method creates a prediction using either the latest version of a model
        or a specific model version.

        Args:
            model_owner (`Optional[str]`): Username of the model's owner.
            model_name (`Optional[str]`): Name of the model.
            version_id (`Optional[str]`): Unique identifier of the model's version.
            params (CreatePredictionParams): Prediction creation params.
            wait (bool): If True, waits until the prediction reaches a terminal state.
                If False, returns the prediction immediately. Defaults to False.
            wait_options (WaitOptions): Options for customizing wait behavior.
                If not provided, default options are used.

        Returns:
            PredictionWithIO: The new prediction.

        Raises:
            TypeError: When an invalid combination of version_id, model_owner,
                and model_name is provided.
        """
        path = self._prediction_path(model_owner, model_name, version_id)

        headers = None
        if "validate_input" in params:
            headers = {"Validate-Input": str(params.pop("validate_input")).lower()}

        prediction = cast(
            PredictionWithIO,
            await self._api.async_request(
                method="POST",
                path=path,
                # The params matches JsonValue at runtime,
                # but static type checkers may not infer this correctly.
                # Hence, we ignore the argument type warning.
                body=params,  # type:ignore[arg-type]
                headers=headers,
            ),
        )

        if wait:
            return await self.async_wait(prediction["id"], options=wait_options)

        return prediction

    def wait(
        self,
        predictionID: str,
        options: WaitOptions = DEFAULT_WAIT_OPTIONS,
    ) -> PredictionWithIO:
        """Wait for a prediction to complete.

        This method uses short polling to check the prediction status at
        regular intervals until it reaches a terminal state. It blocks
        execution and waits until the prediction has finished processing.

        Args:
            predictionID (str): Unique identifier of the prediction.
            options (WaitOptions): Options for customizing wait behavior.
                If not provided, default options are used.

        Returns:
            PredictionWithIO: Prediction after completion.
        """
        while True:
            result = self.get(predictionID)
            if result["status"] in {Status.succeeded, Status.failed}:
                return result
            time.sleep(options.poll_interval)

    async def async_wait(
        self,
        predictionID: str,
        options: WaitOptions = DEFAULT_WAIT_OPTIONS,
    ) -> PredictionWithIO:
        """Wait for a prediction to complete.

        This method uses short polling to check the prediction status at
        regular intervals until it reaches a terminal state. It blocks
        execution and waits until the prediction has finished processing.

        Args:
            predictionID (str): Unique identifier of the prediction.
            options (WaitOptions): Options for customizing wait behavior.
                If not provided, default options are used.

        Returns:
            PredictionWithIO: Prediction after completion.
        """
        while True:
            result = await self.async_get(predictionID)
            if result["status"] in {Status.succeeded, Status.failed}:
                return result
            await asyncio.sleep(options.poll_interval)



"""Hardware module."""

from typing import Optional, TypedDict

from wriftai._resource import Resource
from wriftai.pagination import PaginatedResponse, PaginationOptions


class Hardware(TypedDict):
    """Represents a hardware item."""

    id: str
    """Unique identifier for the hardware."""
    name: str
    """Name of the hardware."""
    gpus: int
    """Number of GPUs available on the hardware."""
    cpus: int
    """Number of CPUs available on the hardware."""
    ram_per_gpu_gb: int
    """Amount of Ram (in GB) allocated per GPU."""
    ram_gb: int
    """Total RAM (in GB) available on the hardware."""
    created_at: str
    """Timestamp when the hardware was created."""


class HardwareResource(Resource):
    """Resource for operations related to hardware."""

    _API_PREFIX = "/hardware"

    def list(
        self, pagination_options: Optional[PaginationOptions] = None
    ) -> PaginatedResponse[Hardware]:
        """List hardware.

        Args:
            pagination_options (Optional[PaginationOptions]): Optional settings
                to control pagination behavior.

        Returns:
            PaginatedResponse[Hardware]: Paginated response containing
                hardware items and navigation metadata.
        """
        response = self._api.request(
            method="GET", params=pagination_options, path=self._API_PREFIX
        )

        # The response will always match the PaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]

    async def async_list(
        self, pagination_options: Optional[PaginationOptions] = None
    ) -> PaginatedResponse[Hardware]:
        """List hardware.

        Args:
            pagination_options (Optional[PaginationOptions]): Optional settings
                to control pagination behavior.

        Returns:
            PaginatedResponse[Hardware]: Paginated response containing
                hardware items and navigation metadata.
        """
        response = await self._api.async_request(
            method="GET", params=pagination_options, path=self._API_PREFIX
        )
        # The response will always match the PaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]



"""Models module."""

from typing import Optional, TypedDict, cast

from wriftai._resource import Resource
from wriftai.common_types import NotRequired, StrEnum, User, Version
from wriftai.pagination import PaginatedResponse, PaginationOptions


class ModelVisibility(StrEnum):
    """Model visibility states."""

    private = "private"
    public = "public"


class Model(TypedDict):
    """Represents a model."""

    id: str
    """The unique identifier of the model."""
    name: str
    """ The name of the model."""
    created_at: str
    """The time when the model was created."""
    visibility: ModelVisibility
    """The visibility of the model."""
    description: str | None
    """Description of the model."""
    updated_at: str | None
    """The time when the model was updated."""
    source_url: str | None
    """Source url from where the model's code can be referenced."""
    license_url: str | None
    """License url where the model's usage is specified."""
    paper_url: str | None
    """Paper url from where research info on the model
            can be found."""
    owner: User
    """The details of the owner of the model."""
    latest_version: Version | None
    """The details of the latest version of the model."""
    hardware_name: str
    """The name of the hardware used by the model."""
    predictions_count: int
    """The total number of predictions created across all versions
         of the model."""


class ModelsSortBy(StrEnum):
    """Enumeration of possible sorting options for querying models."""

    CREATED_AT_ASC = "created_at_asc"
    CREATED_AT_DESC = "created_at_desc"
    PREDICTION_COUNT_DESC = "prediction_count_desc"


class ModelPaginationOptions(PaginationOptions):
    """Pagination options for querying models."""

    sort_by: NotRequired[ModelsSortBy]
    """The sorting criteria."""


class UpdateModelParams(TypedDict):
    """Parameters for updating a model."""

    name: NotRequired[str]
    """The name of the model."""
    description: NotRequired[str | None]
    """Description of the model."""
    visibility: NotRequired[ModelVisibility]
    """The visibility of the model."""
    hardware_name: NotRequired[str]
    """The name of the hardware used by the model."""
    source_url: NotRequired[str | None]
    """Source url from where the model's code can be referenced."""
    license_url: NotRequired[str | None]
    """License url where the model's usage is specified."""
    paper_url: NotRequired[str | None]
    """Paper url from where research info on the model can be
        found."""


class CreateModelParams(TypedDict):
    """Parameters for creating a model."""

    name: str
    """The name of the model."""
    hardware_name: str
    """The name of the hardware used by the model."""
    visibility: NotRequired[ModelVisibility]
    """The visibility of the model."""
    description: NotRequired[str | None]
    """Description of the model."""
    source_url: NotRequired[str | None]
    """Source url from where the model's code can be referenced."""
    license_url: NotRequired[str | None]
    """License url where the model's usage is specified."""
    paper_url: NotRequired[str | None]
    """Paper url from where research info on the model can be
        found."""


class ModelsResource(Resource):
    """Resource for operations related to models."""

    _SEARCH_MODELS_SUFFIX = "/models"

    def delete(self, owner: str, name: str) -> None:
        """Delete a model.

        Args:
            owner (str): Username of the model's owner.
            name (str): Name of the model.
        """
        self._api.request("DELETE", f"{self._MODELS_API_PREFIX}/{owner}/{name}")

    async def async_delete(self, owner: str, name: str) -> None:
        """Delete a model.

        Args:
            owner (str): Username of the model's owner.
            name (str): Name of the model.
        """
        await self._api.async_request(
            "DELETE", f"{self._MODELS_API_PREFIX}/{owner}/{name}"
        )

    def list(
        self,
        pagination_options: Optional[ModelPaginationOptions] = None,
        owner: Optional[str] = None,
    ) -> PaginatedResponse[Model]:
        """List models.

        Args:
            pagination_options (Optional[ModelPaginationOptions]): Optional settings
                to control pagination behavior.
            owner (`Optional[str]`): Username of the model's owner to fetch models for.
                If None, all models are fetched.

        Returns:
            PaginatedResponse[Model]: Paginated response containing models and
                navigation metadata.
        """
        path = self._build_list_path(owner)

        response = self._api.request(method="GET", params=pagination_options, path=path)

        # The response will always match the ModelPaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]

    async def async_list(
        self,
        pagination_options: Optional[ModelPaginationOptions] = None,
        owner: Optional[str] = None,
    ) -> PaginatedResponse[Model]:
        """List models.

        Args:
            pagination_options (`Optional[ModelPaginationOptions]`): Optional settings
                to control pagination behavior.
            owner (`Optional[str]`): Username of the model's owner to fetch models for.
                If None, all models are fetched.

        Returns:
            PaginatedResponse[Model]: Paginated response containing models and
                navigation metadata.
        """
        path = self._build_list_path(owner)

        response = await self._api.async_request(
            method="GET", params=pagination_options, path=path
        )
        # The response will always match the ModelPaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]

    def _build_list_path(self, owner: Optional[str] = None) -> str:
        """Construct the API path for listing models.

        Args:
            owner (Optional[str]): Username of the model's owner to fetch models for.
                If None, returns a path to fetch all models.

        Returns:
            str: The constructed API path for listing models.
        """
        return (
            f"{self._MODELS_API_PREFIX}/{owner}" if owner else self._MODELS_API_PREFIX
        )

    def get(self, owner: str, name: str) -> Model:
        """Get a model.

        Args:
            owner (str): Username of the model's owner.
            name (str): Name of the model.

        Returns:
            Model: A model object.
        """
        response = self._api.request("GET", f"{self._MODELS_API_PREFIX}/{owner}/{name}")

        return cast(Model, response)

    async def async_get(self, owner: str, name: str) -> Model:
        """Get a model.

        Args:
            owner (str): Username of the model's owner.
            name (str): Name of the model.

        Returns:
            Model: A model object.
        """
        response = await self._api.async_request(
            "GET", f"{self._MODELS_API_PREFIX}/{owner}/{name}"
        )

        return cast(Model, response)

    def create(self, params: CreateModelParams) -> Model:
        """Create a model.

        Args:
            params (CreateModelParams): Model creation parameters.

        Returns:
            Model: The new model.
        """
        response = self._api.request(
            method="POST",
            path=self._MODELS_API_PREFIX,
            # The params matches JsonValue at runtime,
            # but static type checkers may not infer this correctly.
            # Hence, we ignore the argument type warning.
            body=params,  # type:ignore[arg-type]
        )

        return cast(Model, response)

    async def async_create(self, params: CreateModelParams) -> Model:
        """Create a model.

        Args:
            params (CreateModelParams): Model creation parameters.

        Returns:
            Model: The new model.
        """
        response = await self._api.async_request(
            method="POST",
            path=self._MODELS_API_PREFIX,
            # The params matches JsonValue at runtime,
            # but static type checkers may not infer this correctly.
            # Hence, we ignore the argument type warning.
            body=params,  # type:ignore[arg-type]
        )

        return cast(Model, response)

    def update(
        self,
        owner: str,
        name: str,
        params: UpdateModelParams,
    ) -> Model:
        """Update a model.

        Args:
            owner (str): Username of the model's owner.
            name (str): Name of the model.
            params (UpdateModelParams): The fields to update.

        Returns:
            Model: The updated model.
        """
        response = self._api.request(
            method="PATCH",
            path=f"{self._MODELS_API_PREFIX}/{owner}/{name}",
            # The params matches JsonValue at runtime,
            # but static type checkers may not infer this correctly.
            # Hence, we ignore the argument type warning.
            body=params,  # type: ignore[arg-type]
        )
        return cast(Model, response)

    async def async_update(
        self,
        owner: str,
        name: str,
        params: UpdateModelParams,
    ) -> Model:
        """Update a model.

        Args:
            owner (str): Username of the model's owner.
            name (str): Name of the model.
            params (UpdateModelParams): The fields to update.

        Returns:
            Model: The updated model.
        """
        response = await self._api.async_request(
            method="PATCH",
            path=f"{self._MODELS_API_PREFIX}/{owner}/{name}",
            # The params matches JsonValue at runtime,
            # but static type checkers may not infer this correctly.
            # Hence, we ignore the argument type warning.
            body=params,  # type: ignore[arg-type]
        )

        return cast(Model, response)

    def search(
        self,
        q: str,
        pagination_options: Optional[PaginationOptions] = None,
    ) -> PaginatedResponse[Model]:
        """Search models.

        Args:
            q (str): The search query.
            pagination_options (Optional[PaginationOptions]): Optional settings to
                control pagination behavior.

        Returns:
            PaginatedResponse[Model]: Paginated response containing models
                and navigation metadata.
        """
        params = self._build_search_params(q, pagination_options)

        response = self._api.request(
            method="GET",
            params=params,
            path=f"{self._SEARCH_API_PREFIX}{self._SEARCH_MODELS_SUFFIX}",
        )
        # The response will always match the PaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]

    async def async_search(
        self,
        q: str,
        pagination_options: Optional[PaginationOptions] = None,
    ) -> PaginatedResponse[Model]:
        """Search models.

        Args:
            q (str): The search query.
            pagination_options (Optional[PaginationOptions]): Optional settings to
                control pagintation behavior.

        Returns:
            PaginatedResponse[Model]: Paginated response containing models
                and navigation metadata.
        """
        params = self._build_search_params(q, pagination_options)

        response = await self._api.async_request(
            method="GET",
            params=params,
            path=f"{self._SEARCH_API_PREFIX}{self._SEARCH_MODELS_SUFFIX}",
        )
        # The response will always match the PaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]



from typing import Callable

import httpx
import pytest
from respx import Router

from tests.integration.conftest import Route
from wriftai._client import Client, ClientOptions
from wriftai.pagination import PaginatedResponse
from wriftai.users import UserPaginationOptions, UsersSortBy


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_get(mock_router: Callable[..., Router], async_flag: bool) -> None:
    test_username = "user_123"
    expected_json = {
        "id": "c12258c4-ed83-4d7b-a784-8ed55412325a",
        "username": test_username,
        "avatar_url": "https://cdn.example.com/avatars/user.png",
        "name": "user",
        "bio": "Software Development Engineer",
        "urls": ["https://user.dev", "https://github.com/user"],
        "location": "Karachi, Sindh",
        "company": "Sych",
        "created_at": "2024-08-13T11:48:44.371093Z",
        "updated_at": "2025-08-13T11:48:44.371103Z",
    }

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1

    router = mock_router(
        route=Route(
            method="GET",
            path=f"/users/{test_username}",
            status_code=200,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        response = await client.users.async_get(username=test_username)
    else:
        response = client.users.get(username=test_username)

    assert response == expected_json
    assert call_count == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_list(mock_router: Callable[..., Router], async_flag: bool) -> None:
    expected_cursor = "abc123"
    expected_page_size = "10"
    expected_json = {
        "items": [
            {
                "id": "c12258c4-ed83-4d7b-a784-8ed55412325a",
                "username": "test_username",
                "avatar_url": "https://cdn.example.com/avatars/user.png",
                "name": "user",
                "bio": "Software Development Engineer",
                "urls": ["https://user.dev", "https://github.com/user"],
                "location": "Karachi, Sindh",
                "company": "Sych",
                "created_at": "2024-08-13T11:48:44.371093Z",
                "updated_at": "2025-08-13T11:48:44.371103Z",
            }
        ],
        "next_cursor": "abc123",
        "previous_cursor": None,
        "next_url": "/users?cursor=abc123",
        "previous_url": None,
    }

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1
        assert request.url.params.get("cursor") == expected_cursor
        assert request.url.params.get("page_size") == expected_page_size
        assert request.url.params.get("sort_by") == UsersSortBy.CREATED_AT_ASC
        assert set(request.url.params.keys()).issubset({
            "cursor",
            "page_size",
            "sort_by",
        })

    router = mock_router(
        route=Route(
            method="GET",
            path="/users",
            status_code=200,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    pagination_options: UserPaginationOptions = {
        "cursor": expected_cursor,
        "page_size": int(expected_page_size),
        "sort_by": UsersSortBy.CREATED_AT_ASC,
    }

    if async_flag:
        response = await client.users.async_list(pagination_options=pagination_options)
    else:
        response = client.users.list(pagination_options=pagination_options)

    # expected_json is a dictionary.
    # Although the structure matches the expected fields, static type checkers like mypy
    # may not be able to verify this due to dynamic typing.
    assert response == PaginatedResponse(**expected_json)  # type:ignore[arg-type]
    assert call_count == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_search(mock_router: Callable[..., Router], async_flag: bool) -> None:
    expected_query = "dummy_query"
    expected_json = {
        "items": [
            {
                "id": "c12258c4-ed83-4d7b-a784-8ed55412325a",
                "username": "Kunal_Kumar",
                "avatar_url": "https://cdn.example.com/avatars/Kunal.png",
                "name": "Kunal",
                "bio": "Software Development Engineer",
                "urls": ["https://Kunal.dev", "https://github.com/Kunal"],
                "location": "Karachi, Sindh",
                "company": "Sych",
                "created_at": "2024-08-13T11:48:44.371093Z",
                "updated_at": "2025-08-13T11:48:44.371103Z",
            }
        ],
        "next_cursor": "abc123",
        "previous_cursor": None,
        "next_url": "/search/users?cursor=abc123",
        "previous_url": None,
    }

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1
        assert request.url.params.get("q") == expected_query
        assert set(request.url.params.keys()).issubset({"q"})

    router = mock_router(
        route=Route(
            method="GET",
            path="/search/users",
            status_code=200,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        response = await client.users.async_search(q=expected_query)
    else:
        response = client.users.search(q=expected_query)

    # expected_json is a dictionary.
    # Although the structure matches the expected fields, static type checkers like mypy
    # may not be able to verify this due to dynamic typing.
    assert response == PaginatedResponse(**expected_json)  # type:ignore[arg-type]
    assert call_count == 1



import json
from typing import Callable

import httpx
import pytest
from respx import Router

from tests.integration.conftest import Route
from wriftai._client import Client, ClientOptions
from wriftai.authenticated_user import UpdateUserParams
from wriftai.models import ModelVisibility
from wriftai.pagination import PaginatedResponse, PaginationOptions


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_get(mock_router: Callable[..., Router], async_flag: bool) -> None:
    expected_json = {
        "id": "c12258c4-ed83-4d7b-a784-8ed55412325a",
        "username": "test_username",
        "avatar_url": "https://cdn.wriftai.com/avatars/user.png",
        "name": "user",
        "bio": "Software Development Engineer",
        "urls": ["https://user.dev", "https://github.com/username"],
        "location": "Karachi, Sindh",
        "company": "WriftAI",
        "created_at": "2025-08-13T11:48:44.371093Z",
        "updated_at": "2025-08-13T11:48:44.371103Z",
    }

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1

    router = mock_router(
        route=Route(
            method="GET",
            path="/user",
            status_code=200,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        response = await client.authenticated_user.async_get()
    else:
        response = client.authenticated_user.get()

    assert dict(response) == expected_json
    assert call_count == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_models_with_pagination_options(
    mock_router: Callable[..., Router], async_flag: bool
) -> None:
    expected_cursor = "abc123"
    expected_page_size = "10"

    expected_json = {
        "items": [
            {
                "id": "123e4567-e89b-12d3-a456-426614174000",
                "name": "test_model",
                "created_at": "2025-08-01T10:00:00Z",
                "visibility": ModelVisibility.public,
                "description": "Test model description",
                "updated_at": "2025-08-15T14:30:00Z",
                "source_url": "https://example.com/source",
                "license_url": "https://example.com/license",
                "paper_url": "https://example.com/paper",
                "owner": {
                    "id": "user-001",
                    "username": "Kunal_Kumar",
                    "avatar_url": "https://example.com/avatar.png",
                    "name": "Kunal Kumar",
                    "bio": "Soft Dev",
                    "urls": ["https://github.com/kunal"],
                    "location": "Karachi",
                    "company": "Sych",
                    "created_at": "2024-05-10T09:00:00Z",
                    "updated_at": "2025-07-20T16:45:00Z",
                },
                "latest_version": {
                    "id": "ver-001",
                    "release_notes": "Initial release.",
                    "created_at": "2025-08-10T09:00:00Z",
                    "schemas": {
                        "prediction": {
                            "input": {"key1": "value1", "key2": 123},
                            "output": {"result": True, "message": "Success"},
                        }
                    },
                    "container_image_digest": "sha256:abc123def456ghi",
                },
                "hardware_name": "NVIDIA A100",
                "predictions_count": 2048,
            }
        ],
        "next_cursor": "abc123",
        "previous_cursor": None,
        "next_url": "/user/models?cursor=abc123",
        "previous_url": None,
    }

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1
        assert request.url.params.get("cursor") == expected_cursor
        assert request.url.params.get("page_size") == expected_page_size
        assert set(request.url.params.keys()).issubset({"cursor", "page_size"})

    router = mock_router(
        route=Route(
            method="GET",
            path="/user/models",
            status_code=200,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    pagination_options = PaginationOptions({
        "cursor": expected_cursor,
        "page_size": int(expected_page_size),
    })

    if async_flag:
        response = await client.authenticated_user.async_models(
            pagination_options=pagination_options
        )
    else:
        response = client.authenticated_user.models(
            pagination_options=pagination_options
        )

    # expected_json is a dictionary.
    # Although the structure matches the expected fields
    # static type checkers like mypy
    # may not be able to verify this due to dynamic typing.
    assert response == PaginatedResponse(**expected_json)  # type: ignore[arg-type]
    assert call_count == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_update(mock_router: Callable[..., Router], async_flag: bool) -> None:
    username = "test_username"
    name = "Updated User"
    company = "Updated Company"
    location = "Updated Location"
    bio = "Updated Bio"
    urls = ["https://user.dev", "https://github.com/username"]

    expected_json = {
        "id": "c12258c4-ed83-4d7b-a784-8ed55412325a",
        "username": username,
        "avatar_url": "https://cdn.wriftai.com/avatars/user.png",
        "name": name,
        "bio": bio,
        "urls": urls,
        "location": location,
        "company": company,
        "created_at": "2025-08-13T11:48:44.371093Z",
        "updated_at": "2025-09-01T12:00:00.000000Z",
    }

    payload: UpdateUserParams = {
        "username": username,
        "name": name,
        "company": company,
        "location": location,
        "bio": bio,
        "urls": urls,
    }

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1
        json_payload = json.loads(request.content.decode())
        assert json_payload == payload

    router = mock_router(
        route=Route(
            method="PATCH",
            path="/user",
            status_code=200,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        response = await client.authenticated_user.async_update(params=payload)
    else:
        response = client.authenticated_user.update(params=payload)

    assert response == expected_json
    assert call_count == 1



from typing import Callable

import httpx
import pytest
from respx import Router

from tests.integration.conftest import Route
from wriftai._client import Client, ClientOptions
from wriftai.pagination import PaginatedResponse, PaginationOptions


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_list(mock_router: Callable[..., Router], async_flag: bool) -> None:
    expected_cursor = "abc123"
    expected_page_size = "10"
    expected_json = {
        "items": [
            {
                "id": "551c7d27-11e8-40a4-90d3-e8d8c22c5894",
                "name": "NVIDIA A100",
                "gpus": 8,
                "cpus": 64,
                "ram_per_gpu_gb": 40,
                "ram_gb": 320,
                "created_at": "2025-08-13T11:48:44.371093Z",
            }
        ],
        "next_cursor": "abc123",
        "previous_cursor": None,
        "next_url": "/hardware?cursor=abc123",
        "previous_url": None,
    }

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1
        assert request.url.params.get("cursor") == expected_cursor
        assert request.url.params.get("page_size") == expected_page_size
        assert set(request.url.params.keys()).issubset({"cursor", "page_size"})

    router = mock_router(
        route=Route(
            method="GET",
            path="/hardware",
            status_code=200,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    pagination_options = PaginationOptions({
        "cursor": expected_cursor,
        "page_size": int(expected_page_size),
    })

    if async_flag:
        response = await client.hardware.async_list(
            pagination_options=pagination_options
        )
    else:
        response = client.hardware.list(pagination_options=pagination_options)

    # expected_json is a dictionary.
    # Although the structure matches the expected fields, static type checkers like mypy
    # may not be able to verify this due to dynamic typing.
    assert response == PaginatedResponse(**expected_json)  # type:ignore[arg-type]
    assert call_count == 1



"""Resource module."""

from abc import ABC
from collections.abc import Mapping
from typing import Any, Optional

from wriftai.api import API
from wriftai.pagination import PaginationOptions


class Resource(ABC):
    """Abstract base class for API resources."""

    _api: API
    _MODELS_API_PREFIX = "/models"
    _VERSIONS_API_PREFIX = "/versions"
    _SEARCH_API_PREFIX = "/search"

    def __init__(self, api: API) -> None:
        """Initializes the Resource with an API instance.

        Args:
            api (API): An instance of the API class.
        """
        self._api = api

    def _build_search_params(
        self, q: str, pagination_options: Optional[PaginationOptions] = None
    ) -> Mapping[str, Any]:
        """Build search parameters.

        Args:
            q (str): The search query.
            pagination_options (Optional[PaginationOptions]): Optional settings to
                control pagination behavior.

        Returns:
            Mapping[str, Any]: Parameters for searching.
        """
        params = {**pagination_options} if pagination_options else {}
        params["q"] = q

        return params



from typing import Any, Callable, Optional, TypedDict

import pytest
import respx
from httpx import Request, Response

from wriftai._client import API_BASE_URL


class Route(TypedDict):
    method: str
    path: str
    status_code: int
    json: dict[str, Any]


@pytest.fixture(scope="function")
def mock_router() -> Callable[..., respx.Router]:
    def _create_mock(
        *,
        route: Route,
        base_url: str = API_BASE_URL,
        callback: Optional[Callable[[Request], None]] = None,
    ) -> respx.Router:
        router = respx.Router(base_url=base_url)

        def handler(request: Request) -> Response:
            if callback:
                callback(request)
            return Response(route.get("status_code", 200), json=route["json"])

        router.route(
            method=route.get("method", "GET"),
            path=route["path"],
        ).mock(side_effect=handler)

        return router

    return _create_mock



"""Versions module."""

from typing import Optional, TypedDict, cast

from wriftai._resource import Resource
from wriftai.common_types import Schemas, Version
from wriftai.pagination import PaginatedResponse, PaginationOptions


class CreateVersionParams(TypedDict):
    """Parameters for creating a version of a model.

    Attributes:
        release_notes (str): Information about changes such as new features,
            bug fixes, or optimizations in this version.
        schemas (Schemas): Schemas for the model version.
        container_image_digest (str): SHA256 hash digest of the version's
            container image.
    """

    release_notes: str
    schemas: Schemas
    container_image_digest: str


class Versions(Resource):
    """Resource for operations related to versions."""

    _VERSIONS_API_SUFFIX = "/versions"

    def get(self, version_id: str) -> Version:
        """Fetch a model version by its id.

        Args:
            version_id(str): The unique identifier of the version.

        Returns:
            Version: The model version.
        """
        response = self._api.request("GET", f"{self._VERSIONS_API_PREFIX}/{version_id}")
        return cast(Version, response)

    async def async_get(self, version_id: str) -> Version:
        """Fetch a model version by its id.

        Args:
            version_id(str): The unique identifier of the version.

        Returns:
            Version: The model version.
        """
        response = await self._api.async_request(
            "GET", f"{self._VERSIONS_API_PREFIX}/{version_id}"
        )
        return cast(Version, response)

    def list(
        self,
        model_owner: str,
        model_name: str,
        pagination_options: Optional[PaginationOptions] = None,
    ) -> PaginatedResponse[Version]:
        """List versions.

        Args:
            model_owner (str): Username of the model's owner.
            model_name (str): Name of the model.
            pagination_options (Optional[PaginationOptions]): Optional settings
                to control pagination behavior.

        Returns:
            PaginatedResponse[Version]: Paginated response containing versions
                and navigation metadata.
        """
        path = (
            f"{self._MODELS_API_PREFIX}/{model_owner}"
            f"/{model_name}{self._VERSIONS_API_SUFFIX}"
        )
        response = self._api.request(method="GET", params=pagination_options, path=path)

        # The response will always match the PaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]

    async def async_list(
        self,
        model_owner: str,
        model_name: str,
        pagination_options: Optional[PaginationOptions] = None,
    ) -> PaginatedResponse[Version]:
        """List versions.

        Args:
            model_owner (str): Username of the model's owner.
            model_name (str): Name of the model.
            pagination_options (Optional[PaginationOptions]): Optional settings
                to control pagination behavior.

        Returns:
            PaginatedResponse[Version]: Paginated response containing versions
                and navigation metadata.
        """
        path = (
            f"{self._MODELS_API_PREFIX}/{model_owner}"
            f"/{model_name}{self._VERSIONS_API_SUFFIX}"
        )
        response = await self._api.async_request(
            method="GET", params=pagination_options, path=path
        )
        # The response will always match the PaginatedResponse structure,
        # but static type checkers may not infer this correctly.
        # Hence, we ignore the argument type warning.
        return PaginatedResponse(**response)  # type:ignore[arg-type]

    def delete(self, version_id: str) -> None:
        """Delete a model version by its id.

        Args:
            version_id(str): The unique identifier of the version.
        """
        self._api.request("DELETE", f"{self._VERSIONS_API_PREFIX}/{version_id}")

    async def async_delete(self, version_id: str) -> None:
        """Delete a model version by its id.

        Args:
            version_id(str): The unique identifier of the version.
        """
        await self._api.async_request(
            "DELETE", f"{self._VERSIONS_API_PREFIX}/{version_id}"
        )

    def create(
        self,
        model_owner: str,
        model_name: str,
        options: CreateVersionParams,
    ) -> Version:
        """Create a version of a model.

        Args:
            model_owner (str): Username of the model's owner.
            model_name (str): Name of the model.
            options (VersionCreateParams): Model's version creation parameters.

        Returns:
            Version: The new version.
        """
        path = (
            f"{self._MODELS_API_PREFIX}/{model_owner}"
            f"/{model_name}{self._VERSIONS_API_SUFFIX}"
        )
        response = self._api.request(
            "POST",
            path,
            # The params matches JsonValue at runtime,
            # but static type checkers may not infer this correctly.
            # Hence, we ignore the argument type warning.
            body=options,  # type:ignore[arg-type]
        )
        return cast(Version, response)

    async def async_create(
        self,
        model_owner: str,
        model_name: str,
        options: CreateVersionParams,
    ) -> Version:
        """Create a version of a model.

        Args:
            model_owner (str): Username of the model's owner.
            model_name (str): Name of the model.
            options (VersionCreateParams): Model's version creation parameters.

        Returns:
            Version: The new version.
        """
        path = (
            f"{self._MODELS_API_PREFIX}/{model_owner}"
            f"/{model_name}{self._VERSIONS_API_SUFFIX}"
        )
        response = await self._api.async_request(
            "POST",
            path,
            # The params matches JsonValue at runtime,
            # but static type checkers may not infer this correctly.
            # Hence, we ignore the argument type warning.
            body=options,  # type:ignore[arg-type]
        )
        return cast(Version, response)



import json
from typing import Callable

import httpx
import pytest
from respx import Router

from tests.integration.conftest import Route
from wriftai._client import Client, ClientOptions
from wriftai.pagination import PaginatedResponse, PaginationOptions
from wriftai.versions import CreateVersionParams


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_get(mock_router: Callable[..., Router], async_flag: bool) -> None:
    version_id = "c12258c4-ed83-4d7b-a784-8ed55412325a"
    expected_json = {
        "id": version_id,
        "release_notes": (
            "Information about changes such as new features,"
            "bug fixes, or optimizations in this version."
        ),
        "created_at": "2025-08-29T11:48:44.371093Z",
        "schemas": {
            "prediction": {
                "input": {"key1": "value1", "key2": 123},
                "output": {"result": True, "message": "Success"},
            }
        },
        "container_image_digest": (
            "94a00394bc5a8ef503fb59db0a7d0ae9e1110866e8aee8ba40cd864cea69ea1a"
        ),
    }

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1

    router = mock_router(
        route=Route(
            method="GET",
            path=f"/versions/{version_id}",
            status_code=200,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        response = await client.versions.async_get(version_id=version_id)
    else:
        response = client.versions.get(version_id=version_id)

    assert response == expected_json
    assert call_count == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_list(mock_router: Callable[..., Router], async_flag: bool) -> None:
    model_owner = "abc"
    model_name = "textgenerator"
    expected_cursor = "abc123"
    expected_page_size = "10"
    expected_json = {
        "items": [
            {
                "id": "c12258c4-ed83-4d7b-a784-8ed55412325a",
                "release_notes": (
                    "Information about changes such as new features,"
                    "bug fixes, or optimizations in this version."
                ),
                "created_at": "2025-08-29T11:48:44.371093Z",
                "schemas": {
                    "prediction": {
                        "input": {"key1": "value1", "key2": 123},
                        "output": {"result": True, "message": "Success"},
                    }
                },
                "container_image_digest": (
                    "94a00394bc5a8ef503fb59db0a7d0ae9e1110866e8aee8ba40cd864cea69ea1a"
                ),
            }
        ],
        "next_cursor": "abc123",
        "previous_cursor": None,
        "next_url": f"/models/{model_owner}/{model_name}/versions?cursor=abc123",
        "previous_url": None,
    }

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1
        assert request.url.params.get("cursor") == expected_cursor
        assert request.url.params.get("page_size") == expected_page_size
        assert set(request.url.params.keys()).issubset({"cursor", "page_size"})

    router = mock_router(
        route=Route(
            method="GET",
            path=f"/models/{model_owner}/{model_name}/versions",
            status_code=200,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    pagination_options = PaginationOptions({
        "cursor": expected_cursor,
        "page_size": int(expected_page_size),
    })

    if async_flag:
        response = await client.versions.async_list(
            model_owner=model_owner,
            model_name=model_name,
            pagination_options=pagination_options,
        )
    else:
        response = client.versions.list(
            model_owner=model_owner,
            model_name=model_name,
            pagination_options=pagination_options,
        )

    # expected_json is a dictionary.
    # Although the structure matches the expected fields, static type checkers like mypy
    # may not be able to verify this due to dynamic typing.
    assert response == PaginatedResponse(**expected_json)  # type:ignore[arg-type]
    assert call_count == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("async_flag", [True, False])
async def test_delete(mock_router: Callable[..., Router], async_flag: bool) -> None:
    version_id = "c12258c4-ed83-4d7b-a784-8ed55412325a"

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1

    router = mock_router(
        route=Route(
            method="DELETE",
            path=f"/versions/{version_id}",
            status_code=204,
            json={},
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        await client.versions.async_delete(version_id=version_id)
    else:
        client.versions.delete(version_id=version_id)
    assert call_count == 1


@pytest.mark.parametrize("async_flag", [True, False])
@pytest.mark.asyncio
async def test_create(mock_router: Callable[..., Router], async_flag: bool) -> None:
    model_owner = "abc"
    model_name = "textgenerator"
    options: CreateVersionParams = {
        "release_notes": "Initial release with basic features",
        "container_image_digest": "sha256:" + "a" * 64,
        "schemas": {
            "prediction": {
                "input": {"key1": "value1", "key2": 123},
                "output": {"result": True, "message": "Success"},
            }
        },
    }

    expected_json = {
        "id": "c12258c4-ed83-4d7b-a784-8ed55412325a",
        "release_notes": options["release_notes"],
        "created_at": "2025-08-29T11:48:44.371093Z",
        "schemas": options["schemas"],
        "container_image_digest": options["container_image_digest"],
    }

    call_count = 0

    def request_assertions(request: httpx.Request) -> None:
        nonlocal call_count
        call_count += 1
        json_payload = json.loads(request.content.decode())
        assert json_payload == options

    router = mock_router(
        route=Route(
            method="POST",
            path=f"/models/{model_owner}/{model_name}/versions",
            status_code=200,
            json=expected_json,
        ),
        callback=request_assertions,
    )

    client = Client(
        client_options=ClientOptions(
            headers={"Authorization": "test-token"},
            timeout=httpx.Timeout(15),
            transport=httpx.MockTransport(router.handler),
        )
    )

    if async_flag:
        response = await client.versions.async_create(
            model_owner=model_owner,
            model_name=model_name,
            options=options,
        )
    else:
        response = client.versions.create(
            model_owner=model_owner,
            model_name=model_name,
            options=options,
        )

    assert response == expected_json
    assert call_count == 1



from unittest.mock import AsyncMock, Mock

import pytest
from httpx import HTTPStatusError

from wriftai.api import API
from wriftai.common_types import JsonValue


def test_request() -> None:
    test_path = "/test-path"
    test_method = "GET"
    test_body: JsonValue = {"key": "value"}
    test_headers = {"test": "test"}
    test_params = {"cursor": "abc123", "page_size": 50}

    mock_sync_client = Mock()
    mock_response = Mock()
    mock_sync_client.request.return_value = mock_response

    api = API(sync_client=mock_sync_client, async_client=Mock())
    result = api.request(
        method=test_method,
        path=test_path,
        body=test_body,
        headers=test_headers,
        params=test_params,
    )

    mock_sync_client.request.assert_called_once_with(
        method=test_method,
        url=test_path,
        json=test_body,
        headers=test_headers,
        params=test_params,
    )
    mock_response.json.assert_called_once_with()

    assert result == mock_response.json.return_value


def test_request_raises_error() -> None:
    test_error_message = "HTTP error"
    mock_sync_client = Mock()
    mock_response = Mock()
    mock_response.raise_for_status.side_effect = HTTPStatusError(
        message=test_error_message, request=Mock(), response=mock_response
    )
    mock_sync_client.request.return_value = mock_response

    api = API(sync_client=mock_sync_client, async_client=Mock())

    with pytest.raises(HTTPStatusError) as e:
        api.request(
            method="GET",
            path="/test-path",
            body={"key": "value"},
            headers={"test": "test"},
            params={"cursor": "abc123", "page_size": 50},
        )

    assert str(e.value) == test_error_message
    mock_response.raise_for_status.assert_called_once_with()


@pytest.mark.asyncio
async def test_async_request() -> None:
    test_path = "/test-path"
    test_method = "GET"
    test_params = {"cursor": "abc123", "page_size": 50}

    mock_async_client = Mock()
    mock_response = Mock()
    mock_async_client.request = AsyncMock(return_value=mock_response)

    api = API(sync_client=Mock(), async_client=mock_async_client)
    result = await api.async_request(
        method=test_method, path=test_path, params=test_params
    )

    mock_async_client.request.assert_awaited_once_with(
        method=test_method, url=test_path, json=None, headers=None, params=test_params
    )
    mock_response.json.assert_called_once_with()

    assert result == mock_response.json.return_value


@pytest.mark.asyncio
async def test_async_request_raises_error() -> None:
    test_error_message = "HTTP error"

    mock_async_client = Mock()
    mock_response = Mock()
    mock_response.raise_for_status.side_effect = HTTPStatusError(
        message=test_error_message, request=Mock(), response=mock_response
    )
    mock_async_client.request = AsyncMock(return_value=mock_response)

    api = API(sync_client=Mock(), async_client=mock_async_client)

    with pytest.raises(HTTPStatusError) as e:
        await api.async_request(
            method="GET",
            path="/test-path",
            body={"key": "value"},
            headers={"test": "test"},
            params={"cursor": "abc123", "page_size": 50},
        )

    assert str(e.value) == test_error_message
    mock_response.raise_for_status.assert_called_once_with()
